# this module exists to resolve circular imports with the events module

from collections import namedtuple
from typing import Optional


_ReferenceKey = namedtuple('_ReferenceKey', 'database schema identifier')


def lowercase(value: Optional[str]) -> Optional[str]:
    if value is None:
        return None
    else:
        return value.lower()


def _make_key(relation) -> _ReferenceKey:
    """Make _ReferenceKeys with lowercase values for the cache so we don't have
    to keep track of quoting
    """
    # databases and schemas can both be None
    return _ReferenceKey(lowercase(relation.database),
                         lowercase(relation.schema),
                         lowercase(relation.identifier))



from dbt.context.context_config import ContextConfig
from dbt.contracts.graph.parsed import ParsedSeedNode
from dbt.node_types import NodeType
from dbt.parser.base import SimpleSQLParser
from dbt.parser.search import FileBlock


class SeedParser(SimpleSQLParser[ParsedSeedNode]):
    def parse_from_dict(self, dct, validate=True) -> ParsedSeedNode:
        if validate:
            ParsedSeedNode.validate(dct)
        return ParsedSeedNode.from_dict(dct)

    @property
    def resource_type(self) -> NodeType:
        return NodeType.Seed

    @classmethod
    def get_compiled_path(cls, block: FileBlock):
        return block.path.relative_path

    def render_with_context(
        self, parsed_node: ParsedSeedNode, config: ContextConfig
    ) -> None:
        """Seeds don't need to do any rendering."""



from dbt import ui
from dbt.node_types import NodeType
from typing import Optional, Union


def format_fancy_output_line(
        msg: str, status: str, index: Optional[int],
        total: Optional[int], execution_time: Optional[float] = None,
        truncate: bool = False
) -> str:
    if index is None or total is None:
        progress = ''
    else:
        progress = '{} of {} '.format(index, total)
    prefix = "{progress}{message}".format(
        progress=progress,
        message=msg)

    truncate_width = ui.printer_width() - 3
    justified = prefix.ljust(ui.printer_width(), ".")
    if truncate and len(justified) > truncate_width:
        justified = justified[:truncate_width] + '...'

    if execution_time is None:
        status_time = ""
    else:
        status_time = " in {execution_time:0.2f}s".format(
            execution_time=execution_time)

    output = "{justified} [{status}{status_time}]".format(
        justified=justified, status=status, status_time=status_time)

    return output


def _pluralize(string: Union[str, NodeType]) -> str:
    try:
        convert = NodeType(string)
    except ValueError:
        return f'{string}s'
    else:
        return convert.pluralize()


def pluralize(count, string: Union[str, NodeType]):
    pluralized: str = str(string)
    if count != 1:
        pluralized = _pluralize(string)
    return f'{count} {pluralized}'



from typing import (
    Any,
    List,
    NamedTuple,
    Optional,
    Dict,
)

# N.B.:
# These stubs were autogenerated by stubgen and then hacked
# to pieces to ensure we had something other than "Any" types
# where using external classes to instantiate event subclasses
# in events/types.py.
#
# This goes away when we turn mypy on for everything.
#
# Don't trust them too much at all!


class _ReferenceKey(NamedTuple):
    database: Any
    schema: Any
    identifier: Any


class _CachedRelation:
    referenced_by: Any
    inner: Any


class BaseRelation:
    path: Any
    type: Optional[Any]
    quote_character: str
    include_policy: Any
    quote_policy: Any
    dbt_created: bool


class InformationSchema(BaseRelation):
    information_schema_view: Optional[str]


class CompiledNode():
    compiled_sql: Optional[str]
    extra_ctes_injected: bool
    extra_ctes: List[Any]
    relation_name: Optional[str]


class CompiledModelNode(CompiledNode):
    resource_type: Any


class ParsedModelNode():
    resource_type: Any


class ParsedHookNode():
    resource_type: Any
    index: Optional[int]


class RunResult():
    status: str
    timing: List[Any]
    thread_id: str
    execution_time: float
    adapter_response: Dict[str, Any]
    message: Optional[str]
    failures: Optional[int]
    node: Any



import functools
import requests
from dbt.events.functions import fire_event
from dbt.events.types import (
    RegistryProgressMakingGETRequest,
    RegistryProgressGETResponse
)
from dbt.utils import memoized, _connection_exception_retry as connection_exception_retry
from dbt import deprecations
import os

if os.getenv('DBT_PACKAGE_HUB_URL'):
    DEFAULT_REGISTRY_BASE_URL = os.getenv('DBT_PACKAGE_HUB_URL')
else:
    DEFAULT_REGISTRY_BASE_URL = 'https://hub.getdbt.com/'


def _get_url(url, registry_base_url=None):
    if registry_base_url is None:
        registry_base_url = DEFAULT_REGISTRY_BASE_URL

    return '{}{}'.format(registry_base_url, url)


def _get_with_retries(path, registry_base_url=None):
    get_fn = functools.partial(_get, path, registry_base_url)
    return connection_exception_retry(get_fn, 5)


def _get(path, registry_base_url=None):
    url = _get_url(path, registry_base_url)
    fire_event(RegistryProgressMakingGETRequest(url=url))
    resp = requests.get(url, timeout=30)
    fire_event(RegistryProgressGETResponse(url=url, resp_code=resp.status_code))
    resp.raise_for_status()

    # It is unexpected for the content of the response to be None so if it is, raising this error
    # will cause this function to retry (if called within _get_with_retries) and hopefully get
    # a response.  This seems to happen when there's an issue with the Hub.
    # See https://github.com/dbt-labs/dbt-core/issues/4577
    if resp.json() is None:
        raise requests.exceptions.ContentDecodingError(
            'Request error: The response is None', response=resp
        )
    return resp.json()


def index(registry_base_url=None):
    return _get_with_retries('api/v1/index.json', registry_base_url)


index_cached = memoized(index)


def packages(registry_base_url=None):
    return _get_with_retries('api/v1/packages.json', registry_base_url)


def package(name, registry_base_url=None):
    response = _get_with_retries('api/v1/{}.json'.format(name), registry_base_url)

    # Either redirectnamespace or redirectname in the JSON response indicate a redirect
    # redirectnamespace redirects based on package ownership
    # redirectname redirects based on package name
    # Both can be present at the same time, or neither. Fails gracefully to old name

    if ('redirectnamespace' in response) or ('redirectname' in response):

        if ('redirectnamespace' in response) and response['redirectnamespace'] is not None:
            use_namespace = response['redirectnamespace']
        else:
            use_namespace = response['namespace']

        if ('redirectname' in response) and response['redirectname'] is not None:
            use_name = response['redirectname']
        else:
            use_name = response['name']

        new_nwo = use_namespace + "/" + use_name
        deprecations.warn('package-redirect', old_name=name, new_name=new_nwo)

    return response


def package_version(name, version, registry_base_url=None):
    return _get_with_retries('api/v1/{}/{}.json'.format(name, version), registry_base_url)


def get_available_versions(name):
    response = package(name)
    return list(response['versions'])



from dataclasses import dataclass
from .types import (
    InfoLevel,
    DebugLevel,
    WarnLevel,
    ErrorLevel,
    ShowException,
    Cli
)


# Keeping log messages for testing separate since they are used for debugging.
# Reuse the existing messages when adding logs to tests.

@dataclass
class IntegrationTestInfo(InfoLevel, Cli):
    msg: str
    code: str = "T001"

    def message(self) -> str:
        return f"Integration Test: {self.msg}"


@dataclass
class IntegrationTestDebug(DebugLevel, Cli):
    msg: str
    code: str = "T002"

    def message(self) -> str:
        return f"Integration Test: {self.msg}"


@dataclass
class IntegrationTestWarn(WarnLevel, Cli):
    msg: str
    code: str = "T003"

    def message(self) -> str:
        return f"Integration Test: {self.msg}"


@dataclass
class IntegrationTestError(ErrorLevel, Cli):
    msg: str
    code: str = "T004"

    def message(self) -> str:
        return f"Integration Test: {self.msg}"


@dataclass
class IntegrationTestException(ShowException, ErrorLevel, Cli):
    msg: str
    code: str = "T005"

    def message(self) -> str:
        return f"Integration Test: {self.msg}"


@dataclass
class UnitTestInfo(InfoLevel, Cli):
    msg: str
    code: str = "T006"

    def message(self) -> str:
        return f"Unit Test: {self.msg}"


# since mypy doesn't run on every file we need to suggest to mypy that every
# class gets instantiated. But we don't actually want to run this code.
# making the conditional `if False` causes mypy to skip it as dead code so
# we need to skirt around that by computing something it doesn't check statically.
#
# TODO remove these lines once we run mypy everywhere.
if 1 == 0:
    IntegrationTestInfo(msg='')
    IntegrationTestDebug(msg='')
    IntegrationTestWarn(msg='')
    IntegrationTestError(msg='')
    IntegrationTestException(msg='')
    UnitTestInfo(msg='')



from .analysis import AnalysisParser  # noqa
from .base import Parser, ConfiguredParser  # noqa
from .singular_test import SingularTestParser  # noqa
from .generic_test import GenericTestParser  # noqa
from .docs import DocumentationParser  # noqa
from .hooks import HookParser  # noqa
from .macros import MacroParser  # noqa
from .models import ModelParser  # noqa
from .schemas import SchemaParser  # noqa
from .seeds import SeedParser  # noqa
from .snapshots import SnapshotParser  # noqa

from . import (  # noqa
    analysis, base, generic_test, singular_test, docs, hooks, macros, models, schemas,
    snapshots
)



from abc import ABCMeta, abstractmethod, abstractproperty
from dataclasses import dataclass
from datetime import datetime
import os
import threading
from typing import Any, Optional

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# These base types define the _required structure_ for the concrete event #
# types defined in types.py                                               #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


# in preparation for #3977
class TestLevel():
    def level_tag(self) -> str:
        return "test"


class DebugLevel():
    def level_tag(self) -> str:
        return "debug"


class InfoLevel():
    def level_tag(self) -> str:
        return "info"


class WarnLevel():
    def level_tag(self) -> str:
        return "warn"


class ErrorLevel():
    def level_tag(self) -> str:
        return "error"


class Cache():
    # Events with this class will only be logged when the `--log-cache-events` flag is passed
    pass


@dataclass
class Node():
    node_path: str
    node_name: str
    unique_id: str
    resource_type: str
    materialized: str
    node_status: str
    node_started_at: datetime
    node_finished_at: Optional[datetime]
    type: str = 'node_status'


@dataclass
class ShowException():
    # N.B.:
    # As long as we stick with the current convention of setting the member vars in the
    # `message` method of subclasses, this is a safe operation.
    # If that ever changes we'll want to reassess.
    def __post_init__(self):
        self.exc_info: Any = True
        self.stack_info: Any = None
        self.extra: Any = None


# TODO add exhaustiveness checking for subclasses
# can't use ABCs with @dataclass because of https://github.com/python/mypy/issues/5374
# top-level superclass for all events
class Event(metaclass=ABCMeta):
    # fields that should be on all events with their default implementations
    log_version: int = 1
    ts: Optional[datetime] = None  # use getter for non-optional
    ts_rfc3339: Optional[str] = None  # use getter for non-optional
    pid: Optional[int] = None  # use getter for non-optional
    node_info: Optional[Node]

    # four digit string code that uniquely identifies this type of event
    # uniqueness and valid characters are enforced by tests
    @abstractproperty
    @staticmethod
    def code() -> str:
        raise Exception("code() not implemented for event")

    # do not define this yourself. inherit it from one of the above level types.
    @abstractmethod
    def level_tag(self) -> str:
        raise Exception("level_tag not implemented for Event")

    # Solely the human readable message. Timestamps and formatting will be added by the logger.
    # Must override yourself
    @abstractmethod
    def message(self) -> str:
        raise Exception("msg not implemented for Event")

    # exactly one time stamp per concrete event
    def get_ts(self) -> datetime:
        if not self.ts:
            self.ts = datetime.utcnow()
            self.ts_rfc3339 = self.ts.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        return self.ts

    # preformatted time stamp
    def get_ts_rfc3339(self) -> str:
        if not self.ts_rfc3339:
            # get_ts() creates the formatted string too so all time logic is centralized
            self.get_ts()
        return self.ts_rfc3339  # type: ignore

    # exactly one pid per concrete event
    def get_pid(self) -> int:
        if not self.pid:
            self.pid = os.getpid()
        return self.pid

    # in theory threads can change so we don't cache them.
    def get_thread_name(self) -> str:
        return threading.current_thread().getName()

    @classmethod
    def get_invocation_id(cls) -> str:
        from dbt.events.functions import get_invocation_id
        return get_invocation_id()

    # default dict factory for all events. can override on concrete classes.
    @classmethod
    def asdict(cls, data: list) -> dict:
        d = dict()
        for k, v in data:
            # stringify all exceptions
            if isinstance(v, Exception) or isinstance(v, BaseException):
                d[k] = str(v)
            # skip all binary data
            elif isinstance(v, bytes):
                continue
            else:
                d[k] = v
        return d


@dataclass  # type: ignore
class NodeInfo(Event, metaclass=ABCMeta):
    report_node_data: Any  # Union[ParsedModelNode, ...] TODO: resolve circular imports

    def get_node_info(self):
        node_info = Node(
            node_path=self.report_node_data.path,
            node_name=self.report_node_data.name,
            unique_id=self.report_node_data.unique_id,
            resource_type=self.report_node_data.resource_type.value,
            materialized=self.report_node_data.config.get('materialized'),
            node_status=str(self.report_node_data._event_status.get('node_status')),
            node_started_at=self.report_node_data._event_status.get("started_at"),
            node_finished_at=self.report_node_data._event_status.get("finished_at")
        )
        return node_info


class File(Event, metaclass=ABCMeta):
    # Solely the human readable message. Timestamps and formatting will be added by the logger.
    def file_msg(self) -> str:
        # returns the event msg unless overriden in the concrete class
        return self.message()


class Cli(Event, metaclass=ABCMeta):
    # Solely the human readable message. Timestamps and formatting will be added by the logger.
    def cli_msg(self) -> str:
        # returns the event msg unless overriden in the concrete class
        return self.message()



import os
import multiprocessing
if os.name != 'nt':
    # https://bugs.python.org/issue41567
    import multiprocessing.popen_spawn_posix  # type: ignore
from pathlib import Path
from typing import Optional

# PROFILES_DIR must be set before the other flags
# It also gets set in main.py and in set_from_args because the rpc server
# doesn't go through exactly the same main arg processing.
DEFAULT_PROFILES_DIR = os.path.join(os.path.expanduser('~'), '.dbt')
PROFILES_DIR = os.path.expanduser(
    os.getenv('DBT_PROFILES_DIR', DEFAULT_PROFILES_DIR)
)

STRICT_MODE = False  # Only here for backwards compatibility
FULL_REFRESH = False  # subcommand
STORE_FAILURES = False  # subcommand

# Global CLI commands
USE_EXPERIMENTAL_PARSER = None
STATIC_PARSER = None
WARN_ERROR = None
WRITE_JSON = None
PARTIAL_PARSE = None
USE_COLORS = None
DEBUG = None
LOG_FORMAT = None
VERSION_CHECK = None
FAIL_FAST = None
SEND_ANONYMOUS_USAGE_STATS = None
PRINTER_WIDTH = 80
WHICH = None
INDIRECT_SELECTION = None
LOG_CACHE_EVENTS = None
EVENT_BUFFER_SIZE = 100000

# Global CLI defaults. These flags are set from three places:
# CLI args, environment variables, and user_config (profiles.yml).
# Environment variables use the pattern 'DBT_{flag name}', like DBT_PROFILES_DIR
flag_defaults = {
    "USE_EXPERIMENTAL_PARSER": False,
    "STATIC_PARSER": True,
    "WARN_ERROR": False,
    "WRITE_JSON": True,
    "PARTIAL_PARSE": True,
    "USE_COLORS": True,
    "PROFILES_DIR": DEFAULT_PROFILES_DIR,
    "DEBUG": False,
    "LOG_FORMAT": None,
    "VERSION_CHECK": True,
    "FAIL_FAST": False,
    "SEND_ANONYMOUS_USAGE_STATS": True,
    "PRINTER_WIDTH": 80,
    "INDIRECT_SELECTION": 'eager',
    "LOG_CACHE_EVENTS": False,
    "EVENT_BUFFER_SIZE": 100000
}


def env_set_truthy(key: str) -> Optional[str]:
    """Return the value if it was set to a "truthy" string value, or None
    otherwise.
    """
    value = os.getenv(key)
    if not value or value.lower() in ('0', 'false', 'f'):
        return None
    return value


def env_set_bool(env_value):
    if env_value in ('1', 't', 'true', 'y', 'yes'):
        return True
    return False


def env_set_path(key: str) -> Optional[Path]:
    value = os.getenv(key)
    if value is None:
        return value
    else:
        return Path(value)


MACRO_DEBUGGING = env_set_truthy('DBT_MACRO_DEBUGGING')
DEFER_MODE = env_set_truthy('DBT_DEFER_TO_STATE')
ARTIFACT_STATE_PATH = env_set_path('DBT_ARTIFACT_STATE_PATH')
ENABLE_LEGACY_LOGGER = env_set_truthy('DBT_ENABLE_LEGACY_LOGGER')


def _get_context():
    # TODO: change this back to use fork() on linux when we have made that safe
    return multiprocessing.get_context('spawn')


# This is not a flag, it's a place to store the lock
MP_CONTEXT = _get_context()


def set_from_args(args, user_config):
    global STRICT_MODE, FULL_REFRESH, WARN_ERROR, \
        USE_EXPERIMENTAL_PARSER, STATIC_PARSER, WRITE_JSON, PARTIAL_PARSE, \
        USE_COLORS, STORE_FAILURES, PROFILES_DIR, DEBUG, LOG_FORMAT, INDIRECT_SELECTION, \
        VERSION_CHECK, FAIL_FAST, SEND_ANONYMOUS_USAGE_STATS, PRINTER_WIDTH, \
        WHICH, LOG_CACHE_EVENTS, EVENT_BUFFER_SIZE

    STRICT_MODE = False  # backwards compatibility
    # cli args without user_config or env var option
    FULL_REFRESH = getattr(args, 'full_refresh', FULL_REFRESH)
    STORE_FAILURES = getattr(args, 'store_failures', STORE_FAILURES)
    WHICH = getattr(args, 'which', WHICH)

    # global cli flags with env var and user_config alternatives
    USE_EXPERIMENTAL_PARSER = get_flag_value('USE_EXPERIMENTAL_PARSER', args, user_config)
    STATIC_PARSER = get_flag_value('STATIC_PARSER', args, user_config)
    WARN_ERROR = get_flag_value('WARN_ERROR', args, user_config)
    WRITE_JSON = get_flag_value('WRITE_JSON', args, user_config)
    PARTIAL_PARSE = get_flag_value('PARTIAL_PARSE', args, user_config)
    USE_COLORS = get_flag_value('USE_COLORS', args, user_config)
    PROFILES_DIR = get_flag_value('PROFILES_DIR', args, user_config)
    DEBUG = get_flag_value('DEBUG', args, user_config)
    LOG_FORMAT = get_flag_value('LOG_FORMAT', args, user_config)
    VERSION_CHECK = get_flag_value('VERSION_CHECK', args, user_config)
    FAIL_FAST = get_flag_value('FAIL_FAST', args, user_config)
    SEND_ANONYMOUS_USAGE_STATS = get_flag_value('SEND_ANONYMOUS_USAGE_STATS', args, user_config)
    PRINTER_WIDTH = get_flag_value('PRINTER_WIDTH', args, user_config)
    INDIRECT_SELECTION = get_flag_value('INDIRECT_SELECTION', args, user_config)
    LOG_CACHE_EVENTS = get_flag_value('LOG_CACHE_EVENTS', args, user_config)
    EVENT_BUFFER_SIZE = get_flag_value('EVENT_BUFFER_SIZE', args, user_config)


def get_flag_value(flag, args, user_config):
    lc_flag = flag.lower()
    flag_value = getattr(args, lc_flag, None)
    if flag_value is None:
        # Environment variables use pattern 'DBT_{flag name}'
        env_flag = f"DBT_{flag}"
        env_value = os.getenv(env_flag)
        if env_value is not None and env_value != '':
            env_value = env_value.lower()
            # non Boolean values
            if flag in [
                'LOG_FORMAT',
                'PRINTER_WIDTH',
                'PROFILES_DIR',
                'INDIRECT_SELECTION',
                'EVENT_BUFFER_SIZE'
            ]:
                flag_value = env_value
            else:
                flag_value = env_set_bool(env_value)
        elif user_config is not None and getattr(user_config, lc_flag, None) is not None:
            flag_value = getattr(user_config, lc_flag)
        else:
            flag_value = flag_defaults[flag]
    if flag in ['PRINTER_WIDTH', 'EVENT_BUFFER_SIZE']:  # must be ints
        flag_value = int(flag_value)
    if flag == 'PROFILES_DIR':
        flag_value = os.path.abspath(flag_value)

    return flag_value


def get_flag_dict():
    return {
        "use_experimental_parser": USE_EXPERIMENTAL_PARSER,
        "static_parser": STATIC_PARSER,
        "warn_error": WARN_ERROR,
        "write_json": WRITE_JSON,
        "partial_parse": PARTIAL_PARSE,
        "use_colors": USE_COLORS,
        "profiles_dir": PROFILES_DIR,
        "debug": DEBUG,
        "log_format": LOG_FORMAT,
        "version_check": VERSION_CHECK,
        "fail_fast": FAIL_FAST,
        "send_anonymous_usage_stats": SEND_ANONYMOUS_USAGE_STATS,
        "printer_width": PRINTER_WIDTH,
        "indirect_selection": INDIRECT_SELECTION,
        "log_cache_events": LOG_CACHE_EVENTS,
        "event_buffer_size": EVENT_BUFFER_SIZE
    }



from typing import Optional

from dbt.clients.yaml_helper import (  # noqa:F401
    yaml, safe_load, Loader, Dumper,
)
from dbt.events.functions import fire_event, get_invocation_id
from dbt.events.types import (
    DisableTracking, SendingEvent, SendEventFailure, FlushEvents,
    FlushEventsFailure, TrackingInitializeFailure
)
from dbt import version as dbt_version
from dbt import flags
from snowplow_tracker import Subject, Tracker, Emitter, logger as sp_logger
from snowplow_tracker import SelfDescribingJson
from datetime import datetime

import logbook
import pytz
import platform
import uuid
import requests
import os

sp_logger.setLevel(100)

COLLECTOR_URL = "fishtownanalytics.sinter-collect.com"
COLLECTOR_PROTOCOL = "https"

INVOCATION_SPEC = 'iglu:com.dbt/invocation/jsonschema/1-0-2'
PLATFORM_SPEC = 'iglu:com.dbt/platform/jsonschema/1-0-0'
RUN_MODEL_SPEC = 'iglu:com.dbt/run_model/jsonschema/1-0-1'
INVOCATION_ENV_SPEC = 'iglu:com.dbt/invocation_env/jsonschema/1-0-0'
PACKAGE_INSTALL_SPEC = 'iglu:com.dbt/package_install/jsonschema/1-0-0'
RPC_REQUEST_SPEC = 'iglu:com.dbt/rpc_request/jsonschema/1-0-1'
DEPRECATION_WARN_SPEC = 'iglu:com.dbt/deprecation_warn/jsonschema/1-0-0'
LOAD_ALL_TIMING_SPEC = 'iglu:com.dbt/load_all_timing/jsonschema/1-0-3'
RESOURCE_COUNTS = 'iglu:com.dbt/resource_counts/jsonschema/1-0-0'
EXPERIMENTAL_PARSER = 'iglu:com.dbt/experimental_parser/jsonschema/1-0-0'
PARTIAL_PARSER = 'iglu:com.dbt/partial_parser/jsonschema/1-0-1'
DBT_INVOCATION_ENV = 'DBT_INVOCATION_ENV'


class TimeoutEmitter(Emitter):
    def __init__(self):
        super().__init__(
            COLLECTOR_URL, protocol=COLLECTOR_PROTOCOL,
            buffer_size=30, on_failure=self.handle_failure,
            method='post',
            # don't set this.
            byte_limit=None,
        )

    @staticmethod
    def handle_failure(num_ok, unsent):
        # num_ok will always be 0, unsent will always be 1 entry long, because
        # the buffer is length 1, so not much to talk about
        fire_event(DisableTracking())
        disable_tracking()

    def _log_request(self, request, payload):
        sp_logger.info(f"Sending {request} request to {self.endpoint}...")
        sp_logger.debug(f"Payload: {payload}")

    def _log_result(self, request, status_code):
        msg = f"{request} request finished with status code: {status_code}"
        if self.is_good_status_code(status_code):
            sp_logger.info(msg)
        else:
            sp_logger.warning(msg)

    def http_post(self, payload):
        self._log_request('POST', payload)

        r = requests.post(
            self.endpoint,
            data=payload,
            headers={'content-type': 'application/json; charset=utf-8'},
            timeout=5.0
        )

        self._log_result('GET', r.status_code)
        return r

    def http_get(self, payload):
        self._log_request('GET', payload)

        r = requests.get(self.endpoint, params=payload, timeout=5.0)

        self._log_result('GET', r.status_code)
        return r


emitter = TimeoutEmitter()
tracker = Tracker(
    emitter, namespace="cf", app_id="dbt",
)


class User:

    def __init__(self, cookie_dir):
        self.do_not_track = True
        self.cookie_dir = cookie_dir

        self.id = None
        self.invocation_id = get_invocation_id()
        self.run_started_at = datetime.now(tz=pytz.utc)

    def state(self):
        return "do not track" if self.do_not_track else "tracking"

    @property
    def cookie_path(self):
        return os.path.join(self.cookie_dir, '.user.yml')

    def initialize(self):
        self.do_not_track = False

        cookie = self.get_cookie()
        self.id = cookie.get('id')

        subject = Subject()
        subject.set_user_id(self.id)
        tracker.set_subject(subject)

    def disable_tracking(self):
        self.do_not_track = True
        self.id = None
        self.cookie_dir = None
        tracker.set_subject(None)

    def set_cookie(self):
        # If the user points dbt to a profile directory which exists AND
        # contains a profiles.yml file, then we can set a cookie. If the
        # specified folder does not exist, or if there is not a profiles.yml
        # file in this folder, then an inconsistent cookie can be used. This
        # will change in every dbt invocation until the user points to a
        # profile dir file which contains a valid profiles.yml file.
        #
        # See: https://github.com/dbt-labs/dbt-core/issues/1645

        user = {"id": str(uuid.uuid4())}

        cookie_path = os.path.abspath(self.cookie_dir)
        profiles_file = os.path.join(cookie_path, 'profiles.yml')
        if os.path.exists(cookie_path) and os.path.exists(profiles_file):
            with open(self.cookie_path, "w") as fh:
                yaml.dump(user, fh)

        return user

    def get_cookie(self):
        if not os.path.isfile(self.cookie_path):
            user = self.set_cookie()
        else:
            with open(self.cookie_path, "r") as fh:
                try:
                    user = safe_load(fh)
                    if user is None:
                        user = self.set_cookie()
                except yaml.reader.ReaderError:
                    user = self.set_cookie()
        return user


active_user: Optional[User] = None


def get_run_type(args):
    return 'regular'


def get_invocation_context(user, config, args):
    # this adapter might not have implemented the type or unique_field properties
    try:
        adapter_type = config.credentials.type
    except Exception:
        adapter_type = None
    try:
        adapter_unique_id = config.credentials.hashed_unique_field()
    except Exception:
        adapter_unique_id = None

    return {
        "project_id": None if config is None else config.hashed_name(),
        "user_id": user.id,
        "invocation_id": get_invocation_id(),

        "command": args.which,
        "options": None,
        "version": str(dbt_version.installed),
        "run_type": get_run_type(args),
        "adapter_type": adapter_type,
        "adapter_unique_id": adapter_unique_id,
    }


def get_invocation_start_context(user, config, args):
    data = get_invocation_context(user, config, args)

    start_data = {
        "progress": "start",
        "result_type": None,
        "result": None
    }

    data.update(start_data)
    return SelfDescribingJson(INVOCATION_SPEC, data)


def get_invocation_end_context(user, config, args, result_type):
    data = get_invocation_context(user, config, args)

    start_data = {
        "progress": "end",
        "result_type": result_type,
        "result": None
    }

    data.update(start_data)
    return SelfDescribingJson(INVOCATION_SPEC, data)


def get_invocation_invalid_context(user, config, args, result_type):
    data = get_invocation_context(user, config, args)

    start_data = {
        "progress": "invalid",
        "result_type": result_type,
        "result": None
    }

    data.update(start_data)
    return SelfDescribingJson(INVOCATION_SPEC, data)


def get_platform_context():
    data = {
        "platform": platform.platform(),
        "python": platform.python_version(),
        "python_version": platform.python_implementation(),
    }

    return SelfDescribingJson(PLATFORM_SPEC, data)


def get_dbt_env_context():
    default = 'manual'

    dbt_invocation_env = os.getenv(DBT_INVOCATION_ENV, default)
    if dbt_invocation_env == '':
        dbt_invocation_env = default

    data = {
        "environment": dbt_invocation_env,
    }

    return SelfDescribingJson(INVOCATION_ENV_SPEC, data)


def track(user, *args, **kwargs):
    if user.do_not_track:
        return
    else:
        fire_event(SendingEvent(kwargs=str(kwargs)))
        try:
            tracker.track_struct_event(*args, **kwargs)
        except Exception:
            fire_event(SendEventFailure())


def track_invocation_start(config=None, args=None):
    context = [
        get_invocation_start_context(active_user, config, args),
        get_platform_context(),
        get_dbt_env_context()
    ]

    track(
        active_user,
        category="dbt",
        action='invocation',
        label='start',
        context=context
    )


def track_project_load(options):
    context = [SelfDescribingJson(LOAD_ALL_TIMING_SPEC, options)]
    assert active_user is not None, \
        'Cannot track project loading time when active user is None'

    track(
        active_user,
        category='dbt',
        action='load_project',
        label=get_invocation_id(),
        context=context
    )


def track_resource_counts(resource_counts):
    context = [SelfDescribingJson(RESOURCE_COUNTS, resource_counts)]
    assert active_user is not None, \
        'Cannot track resource counts when active user is None'

    track(
        active_user,
        category='dbt',
        action='resource_counts',
        label=get_invocation_id(),
        context=context
    )


def track_model_run(options):
    context = [SelfDescribingJson(RUN_MODEL_SPEC, options)]
    assert active_user is not None, \
        'Cannot track model runs when active user is None'

    track(
        active_user,
        category="dbt",
        action='run_model',
        label=get_invocation_id(),
        context=context
    )


def track_rpc_request(options):
    context = [SelfDescribingJson(RPC_REQUEST_SPEC, options)]
    assert active_user is not None, \
        'Cannot track rpc requests when active user is None'

    track(
        active_user,
        category="dbt",
        action='rpc_request',
        label=get_invocation_id(),
        context=context
    )


def track_package_install(config, args, options):
    assert active_user is not None, \
        'Cannot track package installs when active user is None'

    invocation_data = get_invocation_context(active_user, config, args)

    context = [
        SelfDescribingJson(INVOCATION_SPEC, invocation_data),
        SelfDescribingJson(PACKAGE_INSTALL_SPEC, options)
    ]

    track(
        active_user,
        category="dbt",
        action='package',
        label=get_invocation_id(),
        property_='install',
        context=context
    )


def track_deprecation_warn(options):

    assert active_user is not None, \
        'Cannot track deprecation warnings when active user is None'

    context = [
        SelfDescribingJson(DEPRECATION_WARN_SPEC, options)
    ]

    track(
        active_user,
        category="dbt",
        action='deprecation',
        label=get_invocation_id(),
        property_='warn',
        context=context
    )


def track_invocation_end(
        config=None, args=None, result_type=None
):
    user = active_user
    context = [
        get_invocation_end_context(user, config, args, result_type),
        get_platform_context(),
        get_dbt_env_context()
    ]

    assert active_user is not None, \
        'Cannot track invocation end when active user is None'

    track(
        active_user,
        category="dbt",
        action='invocation',
        label='end',
        context=context
    )


def track_invalid_invocation(
        config=None, args=None, result_type=None
):
    assert active_user is not None, \
        'Cannot track invalid invocations when active user is None'

    user = active_user
    invocation_context = get_invocation_invalid_context(
        user,
        config,
        args,
        result_type
    )

    context = [
        invocation_context,
        get_platform_context(),
        get_dbt_env_context()
    ]

    track(
        active_user,
        category="dbt",
        action='invocation',
        label='invalid',
        context=context
    )


def track_experimental_parser_sample(options):
    context = [SelfDescribingJson(EXPERIMENTAL_PARSER, options)]
    assert active_user is not None, \
        'Cannot track experimental parser info when active user is None'

    track(
        active_user,
        category='dbt',
        action='experimental_parser',
        label=get_invocation_id(),
        context=context
    )


def track_partial_parser(options):
    context = [SelfDescribingJson(PARTIAL_PARSER, options)]
    assert active_user is not None, \
        'Cannot track partial parser info when active user is None'

    track(
        active_user,
        category='dbt',
        action='partial_parser',
        label=get_invocation_id(),
        context=context
    )


def flush():
    fire_event(FlushEvents())
    try:
        tracker.flush()
    except Exception:
        fire_event(FlushEventsFailure())


def disable_tracking():
    global active_user
    if active_user is not None:
        active_user.disable_tracking()
    else:
        active_user = User(None)


def do_not_track():
    global active_user
    active_user = User(None)


def initialize_tracking(cookie_dir):
    global active_user
    active_user = User(cookie_dir)
    try:
        active_user.initialize()
    except Exception:
        fire_event(TrackingInitializeFailure())
        active_user = User(None)


class InvocationProcessor(logbook.Processor):
    def __init__(self):
        super().__init__()

    def process(self, record):
        if active_user is not None:
            record.extra.update({
                "run_started_at": active_user.run_started_at.isoformat(),
                "invocation_id": get_invocation_id(),
            })


def initialize_from_flags():
    # Setting these used to be in UserConfig, but had to be moved here
    if flags.SEND_ANONYMOUS_USAGE_STATS:
        initialize_tracking(flags.PROFILES_DIR)
    else:
        do_not_track()



import importlib
import importlib.util
import os
import glob
import json
from typing import Iterator

import requests

import dbt.exceptions
import dbt.semver

from dbt.ui import green, red, yellow
from dbt import flags

PYPI_VERSION_URL = 'https://pypi.org/pypi/dbt-core/json'


def get_latest_version(version_url: str = PYPI_VERSION_URL):
    try:
        resp = requests.get(version_url)
        data = resp.json()
        version_string = data['info']['version']
    except (json.JSONDecodeError, KeyError, requests.RequestException):
        return None

    return dbt.semver.VersionSpecifier.from_version_string(version_string)


def get_installed_version():
    return dbt.semver.VersionSpecifier.from_version_string(__version__)


def get_package_pypi_url(package_name: str) -> str:
    return f'https://pypi.org/pypi/dbt-{package_name}/json'


def get_version_information():
    flags.USE_COLORS = True if not flags.USE_COLORS else None

    installed = get_installed_version()
    latest = get_latest_version()

    installed_s = installed.to_version_string(skip_matcher=True)
    if latest is None:
        latest_s = 'unknown'
    else:
        latest_s = latest.to_version_string(skip_matcher=True)

    version_msg = ("installed version: {}\n"
                   "   latest version: {}\n\n".format(installed_s, latest_s))

    plugin_version_msg = "Plugins:\n"
    for plugin_name, version in _get_dbt_plugins_info():
        plugin_version = dbt.semver.VersionSpecifier.from_version_string(version)
        latest_plugin_version = get_latest_version(version_url=get_package_pypi_url(plugin_name))
        plugin_update_msg = ''
        if installed == plugin_version or (
            latest_plugin_version and plugin_version == latest_plugin_version
        ):
            compatibility_msg = green('Up to date!')
        else:
            if latest_plugin_version:
                if installed.major == plugin_version.major:
                    compatibility_msg = yellow('Update available!')
                else:
                    compatibility_msg = red('Out of date!')
                plugin_update_msg = (
                    "  Your version of dbt-{} is out of date! "
                    "You can find instructions for upgrading here:\n"
                    "  https://docs.getdbt.com/dbt-cli/install/overview\n\n"
                ).format(plugin_name)
            else:
                compatibility_msg = yellow('No PYPI version available')

        plugin_version_msg += (
            "  - {}: {} - {}\n"
            "{}"
        ).format(plugin_name, version, compatibility_msg, plugin_update_msg)

    if latest is None:
        return ("{}The latest version of dbt could not be determined!\n"
                "Make sure that the following URL is accessible:\n{}\n\n{}"
                .format(version_msg, PYPI_VERSION_URL, plugin_version_msg)
                )

    if installed == latest:
        return f"{version_msg}{green('Up to date!')}\n\n{plugin_version_msg}"

    elif installed > latest:
        return ("{}Your version of dbt is ahead of the latest "
                "release!\n\n{}".format(version_msg, plugin_version_msg))

    else:
        return ("{}Your version of dbt is out of date! "
                "You can find instructions for upgrading here:\n"
                "https://docs.getdbt.com/docs/installation\n\n{}"
                .format(version_msg, plugin_version_msg))


def _get_adapter_plugin_names() -> Iterator[str]:
    spec = importlib.util.find_spec('dbt.adapters')
    # If None, then nothing provides an importable 'dbt.adapters', so we will
    # not be reporting plugin versions today
    if spec is None or spec.submodule_search_locations is None:
        return
    for adapters_path in spec.submodule_search_locations:
        version_glob = os.path.join(adapters_path, '*', '__version__.py')
        for version_path in glob.glob(version_glob):
            # the path is like .../dbt/adapters/{plugin_name}/__version__.py
            # except it could be \\ on windows!
            plugin_root, _ = os.path.split(version_path)
            _, plugin_name = os.path.split(plugin_root)
            yield plugin_name


def _get_dbt_plugins_info():
    for plugin_name in _get_adapter_plugin_names():
        if plugin_name == 'core':
            continue
        try:
            mod = importlib.import_module(
                f'dbt.adapters.{plugin_name}.__version__'
            )
        except ImportError:
            # not an adapter
            continue
        yield plugin_name, mod.version


__version__ = "1.0.4"
installed = get_installed_version()



import argparse
from dataclasses import dataclass
from dbt.adapters.reference_keys import _make_key, _ReferenceKey
from dbt.events.stubs import (
    _CachedRelation,
    BaseRelation,
    ParsedHookNode,
    ParsedModelNode,
    RunResult
)
from dbt import ui
from dbt.events.base_types import (
    Cli, Event, File, DebugLevel, InfoLevel, WarnLevel, ErrorLevel, ShowException, NodeInfo, Cache
)
from dbt.events.format import format_fancy_output_line, pluralize
from dbt.node_types import NodeType
from typing import Any, Dict, List, Optional, Set, Tuple, TypeVar


# The classes in this file represent the data necessary to describe a
# particular event to both human readable logs, and machine reliable
# event streams. classes extend superclasses that indicate what
# destinations they are intended for, which mypy uses to enforce
# that the necessary methods are defined.


# Type representing Event and all subclasses of Event
T_Event = TypeVar('T_Event', bound=Event)


# Event codes have prefixes which follow this table
#
# | Code |     Description     |
# |:----:|:-------------------:|
# | A    | Pre-project loading |
# | E    | DB adapter          |
# | I    | Project parsing     |
# | M    | Deps generation     |
# | Q    | Node processing     |
# | W    | Node testing        |
# | Y    | Post processing     |
# | Z    | Misc                |
# | T    | Test only           |
#
# The basic idea is that event codes roughly translate to the natural order of running a dbt task

# TODO: remove ingore when this is fixed:
# https://github.com/python/mypy/issues/5374

@dataclass  # type: ignore
class AdapterEventBase(Cli, File):
    name: str
    base_msg: str
    args: Tuple[Any, ...]

    # instead of having this inherit from one of the level classes
    def level_tag(self) -> str:
        raise Exception("level_tag should never be called on AdapterEventBase")

    def message(self) -> str:
        # this class shouldn't be createable, but we can't make it an ABC because of a mypy bug
        if type(self).__name__ == 'AdapterEventBase':
            raise Exception(
                'attempted to create a message for AdapterEventBase which cannot be created'
            )

        # only apply formatting if there are arguments to format.
        # avoids issues like "dict: {k: v}".format() which results in `KeyError 'k'`
        msg = (
            self.base_msg
            if len(self.args) == 0 else
            self.base_msg.format(*self.args)
        )
        return f"{self.name} adapter: {msg}"


@dataclass
class AdapterEventDebug(DebugLevel, AdapterEventBase, ShowException):
    code: str = "E001"
    pass


@dataclass
class AdapterEventInfo(InfoLevel, AdapterEventBase, ShowException):
    code: str = "E002"
    pass


@dataclass
class AdapterEventWarning(WarnLevel, AdapterEventBase, ShowException):
    code: str = "E003"
    pass


@dataclass
class AdapterEventError(ErrorLevel, AdapterEventBase, ShowException):
    code: str = "E004"
    pass


@dataclass
class MainKeyboardInterrupt(InfoLevel, Cli):
    code: str = "Z001"

    def message(self) -> str:
        return "ctrl-c"


# will log to a file if the file logger is configured
@dataclass
class MainEncounteredError(ErrorLevel, Cli):
    e: BaseException
    code: str = "Z002"

    def message(self) -> str:
        return f"Encountered an error:\n{str(self.e)}"


@dataclass
class MainStackTrace(DebugLevel, Cli):
    stack_trace: str
    code: str = "Z003"

    def message(self) -> str:
        return self.stack_trace


@dataclass
class MainReportVersion(InfoLevel, Cli, File):
    v: str  # could be VersionSpecifier instead if we resolved some circular imports
    code: str = "A001"

    def message(self):
        return f"Running with dbt{self.v}"


@dataclass
class MainReportArgs(DebugLevel, Cli, File):
    args: argparse.Namespace
    code: str = "A002"

    def message(self):
        return f"running dbt with arguments {str(self.args)}"

    @classmethod
    def asdict(cls, data: list) -> dict:
        return dict((k, str(v)) for k, v in data)


@dataclass
class MainTrackingUserState(DebugLevel, Cli, File):
    user_state: str
    code: str = "A003"

    def message(self):
        return f"Tracking: {self.user_state}"


@dataclass
class ParsingStart(InfoLevel, Cli, File):
    code: str = "I001"

    def message(self) -> str:
        return "Start parsing."


@dataclass
class ParsingCompiling(InfoLevel, Cli, File):
    code: str = "I002"

    def message(self) -> str:
        return "Compiling."


@dataclass
class ParsingWritingManifest(InfoLevel, Cli, File):
    code: str = "I003"

    def message(self) -> str:
        return "Writing manifest."


@dataclass
class ParsingDone(InfoLevel, Cli, File):
    code: str = "I004"

    def message(self) -> str:
        return "Done."


@dataclass
class ManifestDependenciesLoaded(InfoLevel, Cli, File):
    code: str = "I005"

    def message(self) -> str:
        return "Dependencies loaded"


@dataclass
class ManifestLoaderCreated(InfoLevel, Cli, File):
    code: str = "I006"

    def message(self) -> str:
        return "ManifestLoader created"


@dataclass
class ManifestLoaded(InfoLevel, Cli, File):
    code: str = "I007"

    def message(self) -> str:
        return "Manifest loaded"


@dataclass
class ManifestChecked(InfoLevel, Cli, File):
    code: str = "I008"

    def message(self) -> str:
        return "Manifest checked"


@dataclass
class ManifestFlatGraphBuilt(InfoLevel, Cli, File):
    code: str = "I009"

    def message(self) -> str:
        return "Flat graph built"


@dataclass
class ReportPerformancePath(InfoLevel, Cli, File):
    path: str
    code: str = "I010"

    def message(self) -> str:
        return f"Performance info: {self.path}"


@dataclass
class GitSparseCheckoutSubdirectory(DebugLevel, Cli, File):
    subdir: str
    code: str = "M001"

    def message(self) -> str:
        return f"  Subdirectory specified: {self.subdir}, using sparse checkout."


@dataclass
class GitProgressCheckoutRevision(DebugLevel, Cli, File):
    revision: str
    code: str = "M002"

    def message(self) -> str:
        return f"  Checking out revision {self.revision}."


@dataclass
class GitProgressUpdatingExistingDependency(DebugLevel, Cli, File):
    dir: str
    code: str = "M003"

    def message(self) -> str:
        return f"Updating existing dependency {self.dir}."


@dataclass
class GitProgressPullingNewDependency(DebugLevel, Cli, File):
    dir: str
    code: str = "M004"

    def message(self) -> str:
        return f"Pulling new dependency {self.dir}."


@dataclass
class GitNothingToDo(DebugLevel, Cli, File):
    sha: str
    code: str = "M005"

    def message(self) -> str:
        return f"Already at {self.sha}, nothing to do."


@dataclass
class GitProgressUpdatedCheckoutRange(DebugLevel, Cli, File):
    start_sha: str
    end_sha: str
    code: str = "M006"

    def message(self) -> str:
        return f"  Updated checkout from {self.start_sha} to {self.end_sha}."


@dataclass
class GitProgressCheckedOutAt(DebugLevel, Cli, File):
    end_sha: str
    code: str = "M007"

    def message(self) -> str:
        return f"  Checked out at {self.end_sha}."


@dataclass
class RegistryProgressMakingGETRequest(DebugLevel, Cli, File):
    url: str
    code: str = "M008"

    def message(self) -> str:
        return f"Making package registry request: GET {self.url}"


@dataclass
class RegistryProgressGETResponse(DebugLevel, Cli, File):
    url: str
    resp_code: int
    code: str = "M009"

    def message(self) -> str:
        return f"Response from registry: GET {self.url} {self.resp_code}"


# TODO this was actually `logger.exception(...)` not `logger.error(...)`
@dataclass
class SystemErrorRetrievingModTime(ErrorLevel, Cli, File):
    path: str
    code: str = "Z004"

    def message(self) -> str:
        return f"Error retrieving modification time for file {self.path}"


@dataclass
class SystemCouldNotWrite(DebugLevel, Cli, File):
    path: str
    reason: str
    exc: Exception
    code: str = "Z005"

    def message(self) -> str:
        return (
            f"Could not write to path {self.path}({len(self.path)} characters): "
            f"{self.reason}\nexception: {self.exc}"
        )


@dataclass
class SystemExecutingCmd(DebugLevel, Cli, File):
    cmd: List[str]
    code: str = "Z006"

    def message(self) -> str:
        return f'Executing "{" ".join(self.cmd)}"'


@dataclass
class SystemStdOutMsg(DebugLevel, Cli, File):
    bmsg: bytes
    code: str = "Z007"

    def message(self) -> str:
        return f'STDOUT: "{str(self.bmsg)}"'


@dataclass
class SystemStdErrMsg(DebugLevel, Cli, File):
    bmsg: bytes
    code: str = "Z008"

    def message(self) -> str:
        return f'STDERR: "{str(self.bmsg)}"'


@dataclass
class SystemReportReturnCode(DebugLevel, Cli, File):
    returncode: int
    code: str = "Z009"

    def message(self) -> str:
        return f"command return code={self.returncode}"


@dataclass
class SelectorReportInvalidSelector(InfoLevel, Cli, File):
    selector_methods: dict
    spec_method: str
    raw_spec: str
    code: str = "M010"

    def message(self) -> str:
        valid_selectors = ", ".join(self.selector_methods)
        return (
            f"The '{self.spec_method}' selector specified in {self.raw_spec} is "
            f"invalid. Must be one of [{valid_selectors}]"
        )


@dataclass
class MacroEventInfo(InfoLevel, Cli, File):
    msg: str
    code: str = "M011"

    def message(self) -> str:
        return self.msg


@dataclass
class MacroEventDebug(DebugLevel, Cli, File):
    msg: str
    code: str = "M012"

    def message(self) -> str:
        return self.msg


@dataclass
class NewConnection(DebugLevel, Cli, File):
    conn_type: str
    conn_name: str
    code: str = "E005"

    def message(self) -> str:
        return f'Acquiring new {self.conn_type} connection "{self.conn_name}"'


@dataclass
class ConnectionReused(DebugLevel, Cli, File):
    conn_name: str
    code: str = "E006"

    def message(self) -> str:
        return f"Re-using an available connection from the pool (formerly {self.conn_name})"


@dataclass
class ConnectionLeftOpen(DebugLevel, Cli, File):
    conn_name: Optional[str]
    code: str = "E007"

    def message(self) -> str:
        return f"Connection '{self.conn_name}' was left open."


@dataclass
class ConnectionClosed(DebugLevel, Cli, File):
    conn_name: Optional[str]
    code: str = "E008"

    def message(self) -> str:
        return f"Connection '{self.conn_name}' was properly closed."


@dataclass
class RollbackFailed(ShowException, DebugLevel, Cli, File):
    conn_name: Optional[str]
    code: str = "E009"

    def message(self) -> str:
        return f"Failed to rollback '{self.conn_name}'"


# TODO: can we combine this with ConnectionClosed?
@dataclass
class ConnectionClosed2(DebugLevel, Cli, File):
    conn_name: Optional[str]
    code: str = "E010"

    def message(self) -> str:
        return f"On {self.conn_name}: Close"


# TODO: can we combine this with ConnectionLeftOpen?
@dataclass
class ConnectionLeftOpen2(DebugLevel, Cli, File):
    conn_name: Optional[str]
    code: str = "E011"

    def message(self) -> str:
        return f"On {self.conn_name}: No close available on handle"


@dataclass
class Rollback(DebugLevel, Cli, File):
    conn_name: Optional[str]
    code: str = "E012"

    def message(self) -> str:
        return f"On {self.conn_name}: ROLLBACK"


@dataclass
class CacheMiss(DebugLevel, Cli, File):
    conn_name: str
    database: Optional[str]
    schema: str
    code: str = "E013"

    def message(self) -> str:
        return (
            f'On "{self.conn_name}": cache miss for schema '
            '"{self.database}.{self.schema}", this is inefficient'
        )


@dataclass
class ListRelations(DebugLevel, Cli, File):
    database: Optional[str]
    schema: str
    relations: List[_ReferenceKey]
    code: str = "E014"

    def message(self) -> str:
        return f"with database={self.database}, schema={self.schema}, relations={self.relations}"

    @classmethod
    def asdict(cls, data: list) -> dict:
        d = dict()
        for k, v in data:
            if type(v) == list:
                d[k] = [str(x) for x in v]
        return d


@dataclass
class ConnectionUsed(DebugLevel, Cli, File):
    conn_type: str
    conn_name: Optional[str]
    code: str = "E015"

    def message(self) -> str:
        return f'Using {self.conn_type} connection "{self.conn_name}"'


@dataclass
class SQLQuery(DebugLevel, Cli, File):
    conn_name: Optional[str]
    sql: str
    code: str = "E016"

    def message(self) -> str:
        return f"On {self.conn_name}: {self.sql}"


@dataclass
class SQLQueryStatus(DebugLevel, Cli, File):
    status: str
    elapsed: float
    code: str = "E017"

    def message(self) -> str:
        return f"SQL status: {self.status} in {self.elapsed} seconds"


@dataclass
class SQLCommit(DebugLevel, Cli, File):
    conn_name: str
    code: str = "E018"

    def message(self) -> str:
        return f"On {self.conn_name}: COMMIT"


@dataclass
class ColTypeChange(DebugLevel, Cli, File):
    orig_type: str
    new_type: str
    table: str
    code: str = "E019"

    def message(self) -> str:
        return f"Changing col type from {self.orig_type} to {self.new_type} in table {self.table}"


@dataclass
class SchemaCreation(DebugLevel, Cli, File):
    relation: _ReferenceKey
    code: str = "E020"

    def message(self) -> str:
        return f'Creating schema "{self.relation}"'


@dataclass
class SchemaDrop(DebugLevel, Cli, File):
    relation: _ReferenceKey
    code: str = "E021"

    def message(self) -> str:
        return f'Dropping schema "{self.relation}".'

    @classmethod
    def asdict(cls, data: list) -> dict:
        return dict((k, str(v)) for k, v in data)


# TODO pretty sure this is only ever called in dead code
# see: core/dbt/adapters/cache.py _add_link vs add_link
@dataclass
class UncachedRelation(DebugLevel, Cli, File, Cache):
    dep_key: _ReferenceKey
    ref_key: _ReferenceKey
    code: str = "E022"

    def message(self) -> str:
        return (
            f"{self.dep_key} references {str(self.ref_key)} "
            "but {self.ref_key.database}.{self.ref_key.schema}"
            "is not in the cache, skipping assumed external relation"
        )


@dataclass
class AddLink(DebugLevel, Cli, File, Cache):
    dep_key: _ReferenceKey
    ref_key: _ReferenceKey
    code: str = "E023"

    def message(self) -> str:
        return f"adding link, {self.dep_key} references {self.ref_key}"


@dataclass
class AddRelation(DebugLevel, Cli, File, Cache):
    relation: _ReferenceKey
    code: str = "E024"

    def message(self) -> str:
        return f"Adding relation: {str(self.relation)}"


@dataclass
class DropMissingRelation(DebugLevel, Cli, File, Cache):
    relation: _ReferenceKey
    code: str = "E025"

    def message(self) -> str:
        return f"dropped a nonexistent relationship: {str(self.relation)}"


@dataclass
class DropCascade(DebugLevel, Cli, File, Cache):
    dropped: _ReferenceKey
    consequences: Set[_ReferenceKey]
    code: str = "E026"

    def message(self) -> str:
        return f"drop {self.dropped} is cascading to {self.consequences}"

    @classmethod
    def asdict(cls, data: list) -> dict:
        d = dict()
        for k, v in data:
            if isinstance(v, list):
                d[k] = [str(x) for x in v]
            else:
                d[k] = str(v)  # type: ignore
        return d


@dataclass
class DropRelation(DebugLevel, Cli, File, Cache):
    dropped: _ReferenceKey
    code: str = "E027"

    def message(self) -> str:
        return f"Dropping relation: {self.dropped}"


@dataclass
class UpdateReference(DebugLevel, Cli, File, Cache):
    old_key: _ReferenceKey
    new_key: _ReferenceKey
    cached_key: _ReferenceKey
    code: str = "E028"

    def message(self) -> str:
        return f"updated reference from {self.old_key} -> {self.cached_key} to "\
            "{self.new_key} -> {self.cached_key}"


@dataclass
class TemporaryRelation(DebugLevel, Cli, File, Cache):
    key: _ReferenceKey
    code: str = "E029"

    def message(self) -> str:
        return f"old key {self.key} not found in self.relations, assuming temporary"


@dataclass
class RenameSchema(DebugLevel, Cli, File, Cache):
    old_key: _ReferenceKey
    new_key: _ReferenceKey
    code: str = "E030"

    def message(self) -> str:
        return f"Renaming relation {self.old_key} to {self.new_key}"


@dataclass
class DumpBeforeAddGraph(DebugLevel, Cli, File, Cache):
    # large value. delay creation with fire_event_if.
    dump: Dict[str, List[str]]
    code: str = "E031"

    def message(self) -> str:
        return f"before adding : {self.dump}"


@dataclass
class DumpAfterAddGraph(DebugLevel, Cli, File, Cache):
    # large value. delay creation with fire_event_if.
    dump: Dict[str, List[str]]
    code: str = "E032"

    def message(self) -> str:
        return f"after adding: {self.dump}"


@dataclass
class DumpBeforeRenameSchema(DebugLevel, Cli, File, Cache):
    # large value. delay creation with fire_event_if.
    dump: Dict[str, List[str]]
    code: str = "E033"

    def message(self) -> str:
        return f"before rename: {self.dump}"


@dataclass
class DumpAfterRenameSchema(DebugLevel, Cli, File, Cache):
    # large value. delay creation with fire_event_if.
    dump: Dict[str, List[str]]
    code: str = "E034"

    def message(self) -> str:
        return f"after rename: {self.dump}"


@dataclass
class AdapterImportError(InfoLevel, Cli, File):
    exc: ModuleNotFoundError
    code: str = "E035"

    def message(self) -> str:
        return f"Error importing adapter: {self.exc}"

    @classmethod
    def asdict(cls, data: list) -> dict:
        return dict((k, str(v)) for k, v in data)


@dataclass
class PluginLoadError(ShowException, DebugLevel, Cli, File):
    code: str = "E036"

    def message(self):
        pass


@dataclass
class NewConnectionOpening(DebugLevel, Cli, File):
    connection_state: str
    code: str = "E037"

    def message(self) -> str:
        return f"Opening a new connection, currently in state {self.connection_state}"


@dataclass
class TimingInfoCollected(DebugLevel, Cli, File):
    code: str = "Z010"

    def message(self) -> str:
        return "finished collecting timing info"


@dataclass
class MergedFromState(DebugLevel, Cli, File):
    nbr_merged: int
    sample: List
    code: str = "A004"

    def message(self) -> str:
        return f"Merged {self.nbr_merged} items from state (sample: {self.sample})"


@dataclass
class MissingProfileTarget(InfoLevel, Cli, File):
    profile_name: str
    target_name: str
    code: str = "A005"

    def message(self) -> str:
        return f"target not specified in profile '{self.profile_name}', using '{self.target_name}'"


@dataclass
class InvalidVarsYAML(ErrorLevel, Cli, File):
    code: str = "A008"

    def message(self) -> str:
        return "The YAML provided in the --vars argument is not valid."


@dataclass
class GenericTestFileParse(DebugLevel, Cli, File):
    path: str
    code: str = "I011"

    def message(self) -> str:
        return f"Parsing {self.path}"


@dataclass
class MacroFileParse(DebugLevel, Cli, File):
    path: str
    code: str = "I012"

    def message(self) -> str:
        return f"Parsing {self.path}"


@dataclass
class PartialParsingFullReparseBecauseOfError(InfoLevel, Cli, File):
    code: str = "I013"

    def message(self) -> str:
        return "Partial parsing enabled but an error occurred. Switching to a full re-parse."


@dataclass
class PartialParsingExceptionFile(DebugLevel, Cli, File):
    file: str
    code: str = "I014"

    def message(self) -> str:
        return f"Partial parsing exception processing file {self.file}"


@dataclass
class PartialParsingFile(DebugLevel, Cli, File):
    file_dict: Dict
    code: str = "I015"

    def message(self) -> str:
        return f"PP file: {self.file_dict}"


@dataclass
class PartialParsingException(DebugLevel, Cli, File):
    exc_info: Dict
    code: str = "I016"

    def message(self) -> str:
        return f"PP exception info: {self.exc_info}"


@dataclass
class PartialParsingSkipParsing(DebugLevel, Cli, File):
    code: str = "I017"

    def message(self) -> str:
        return "Partial parsing enabled, no changes found, skipping parsing"


@dataclass
class PartialParsingMacroChangeStartFullParse(InfoLevel, Cli, File):
    code: str = "I018"

    def message(self) -> str:
        return "Change detected to override macro used during parsing. Starting full parse."


@dataclass
class PartialParsingProjectEnvVarsChanged(InfoLevel, Cli, File):
    code: str = "I019"

    def message(self) -> str:
        return "Unable to do partial parsing because env vars used in dbt_project.yml have changed"


@dataclass
class PartialParsingProfileEnvVarsChanged(InfoLevel, Cli, File):
    code: str = "I020"

    def message(self) -> str:
        return "Unable to do partial parsing because env vars used in profiles.yml have changed"


@dataclass
class PartialParsingDeletedMetric(DebugLevel, Cli, File):
    id: str
    code: str = "I021"

    def message(self) -> str:
        return f"Partial parsing: deleted metric {self.id}"


@dataclass
class ManifestWrongMetadataVersion(DebugLevel, Cli, File):
    version: str
    code: str = "I022"

    def message(self) -> str:
        return ("Manifest metadata did not contain correct version. "
                f"Contained '{self.version}' instead.")


@dataclass
class PartialParsingVersionMismatch(InfoLevel, Cli, File):
    saved_version: str
    current_version: str
    code: str = "I023"

    def message(self) -> str:
        return ("Unable to do partial parsing because of a dbt version mismatch. "
                f"Saved manifest version: {self.saved_version}. "
                f"Current version: {self.current_version}.")


@dataclass
class PartialParsingFailedBecauseConfigChange(InfoLevel, Cli, File):
    code: str = "I024"

    def message(self) -> str:
        return ("Unable to do partial parsing because config vars, "
                "config profile, or config target have changed")


@dataclass
class PartialParsingFailedBecauseProfileChange(InfoLevel, Cli, File):
    code: str = "I025"

    def message(self) -> str:
        return ("Unable to do partial parsing because profile has changed")


@dataclass
class PartialParsingFailedBecauseNewProjectDependency(InfoLevel, Cli, File):
    code: str = "I026"

    def message(self) -> str:
        return ("Unable to do partial parsing because a project dependency has been added")


@dataclass
class PartialParsingFailedBecauseHashChanged(InfoLevel, Cli, File):
    code: str = "I027"

    def message(self) -> str:
        return ("Unable to do partial parsing because a project config has changed")


@dataclass
class PartialParsingNotEnabled(DebugLevel, Cli, File):
    code: str = "I028"

    def message(self) -> str:
        return ("Partial parsing not enabled")


@dataclass
class ParsedFileLoadFailed(ShowException, DebugLevel, Cli, File):
    path: str
    exc: Exception
    code: str = "I029"

    def message(self) -> str:
        return f"Failed to load parsed file from disk at {self.path}: {self.exc}"


@dataclass
class PartialParseSaveFileNotFound(InfoLevel, Cli, File):
    code: str = "I030"

    def message(self) -> str:
        return ("Partial parse save file not found. Starting full parse.")


@dataclass
class StaticParserCausedJinjaRendering(DebugLevel, Cli, File):
    path: str
    code: str = "I031"

    def message(self) -> str:
        return f"1605: jinja rendering because of STATIC_PARSER flag. file: {self.path}"


# TODO: Experimental/static parser uses these for testing and some may be a good use case for
#       the `TestLevel` logger once we implement it.  Some will probably stay `DebugLevel`.
@dataclass
class UsingExperimentalParser(DebugLevel, Cli, File):
    path: str
    code: str = "I032"

    def message(self) -> str:
        return f"1610: conducting experimental parser sample on {self.path}"


@dataclass
class SampleFullJinjaRendering(DebugLevel, Cli, File):
    path: str
    code: str = "I033"

    def message(self) -> str:
        return f"1611: conducting full jinja rendering sample on {self.path}"


@dataclass
class StaticParserFallbackJinjaRendering(DebugLevel, Cli, File):
    path: str
    code: str = "I034"

    def message(self) -> str:
        return f"1602: parser fallback to jinja rendering on {self.path}"


@dataclass
class StaticParsingMacroOverrideDetected(DebugLevel, Cli, File):
    path: str
    code: str = "I035"

    def message(self) -> str:
        return f"1601: detected macro override of ref/source/config in the scope of {self.path}"


@dataclass
class StaticParserSuccess(DebugLevel, Cli, File):
    path: str
    code: str = "I036"

    def message(self) -> str:
        return f"1699: static parser successfully parsed {self.path}"


@dataclass
class StaticParserFailure(DebugLevel, Cli, File):
    path: str
    code: str = "I037"

    def message(self) -> str:
        return f"1603: static parser failed on {self.path}"


@dataclass
class ExperimentalParserSuccess(DebugLevel, Cli, File):
    path: str
    code: str = "I038"

    def message(self) -> str:
        return f"1698: experimental parser successfully parsed {self.path}"


@dataclass
class ExperimentalParserFailure(DebugLevel, Cli, File):
    path: str
    code: str = "I039"

    def message(self) -> str:
        return f"1604: experimental parser failed on {self.path}"


@dataclass
class PartialParsingEnabled(DebugLevel, Cli, File):
    deleted: int
    added: int
    changed: int
    code: str = "I040"

    def message(self) -> str:
        return (f"Partial parsing enabled: "
                f"{self.deleted} files deleted, "
                f"{self.added} files added, "
                f"{self.changed} files changed.")


@dataclass
class PartialParsingAddedFile(DebugLevel, Cli, File):
    file_id: str
    code: str = "I041"

    def message(self) -> str:
        return f"Partial parsing: added file: {self.file_id}"


@dataclass
class PartialParsingDeletedFile(DebugLevel, Cli, File):
    file_id: str
    code: str = "I042"

    def message(self) -> str:
        return f"Partial parsing: deleted file: {self.file_id}"


@dataclass
class PartialParsingUpdatedFile(DebugLevel, Cli, File):
    file_id: str
    code: str = "I043"

    def message(self) -> str:
        return f"Partial parsing: updated file: {self.file_id}"


@dataclass
class PartialParsingNodeMissingInSourceFile(DebugLevel, Cli, File):
    source_file: str
    code: str = "I044"

    def message(self) -> str:
        return f"Partial parsing: node not found for source_file {self.source_file}"


@dataclass
class PartialParsingMissingNodes(DebugLevel, Cli, File):
    file_id: str
    code: str = "I045"

    def message(self) -> str:
        return f"No nodes found for source file {self.file_id}"


@dataclass
class PartialParsingChildMapMissingUniqueID(DebugLevel, Cli, File):
    unique_id: str
    code: str = "I046"

    def message(self) -> str:
        return f"Partial parsing: {self.unique_id} not found in child_map"


@dataclass
class PartialParsingUpdateSchemaFile(DebugLevel, Cli, File):
    file_id: str
    code: str = "I047"

    def message(self) -> str:
        return f"Partial parsing: update schema file: {self.file_id}"


@dataclass
class PartialParsingDeletedSource(DebugLevel, Cli, File):
    unique_id: str
    code: str = "I048"

    def message(self) -> str:
        return f"Partial parsing: deleted source {self.unique_id}"


@dataclass
class PartialParsingDeletedExposure(DebugLevel, Cli, File):
    unique_id: str
    code: str = "I049"

    def message(self) -> str:
        return f"Partial parsing: deleted exposure {self.unique_id}"


@dataclass
class InvalidDisabledSourceInTestNode(WarnLevel, Cli, File):
    msg: str
    code: str = "I050"

    def message(self) -> str:
        return ui.warning_tag(self.msg)


@dataclass
class InvalidRefInTestNode(DebugLevel, Cli, File):
    msg: str
    code: str = "I051"

    def message(self) -> str:
        return self.msg


@dataclass
class RunningOperationCaughtError(ErrorLevel, Cli, File):
    exc: Exception
    code: str = "Q001"

    def message(self) -> str:
        return f'Encountered an error while running operation: {self.exc}'


@dataclass
class RunningOperationUncaughtError(ErrorLevel, Cli, File):
    exc: Exception
    code: str = "FF01"

    def message(self) -> str:
        return f'Encountered an error while running operation: {self.exc}'


@dataclass
class DbtProjectError(ErrorLevel, Cli, File):
    code: str = "A009"

    def message(self) -> str:
        return "Encountered an error while reading the project:"


@dataclass
class DbtProjectErrorException(ErrorLevel, Cli, File):
    exc: Exception
    code: str = "A010"

    def message(self) -> str:
        return f"  ERROR: {str(self.exc)}"


@dataclass
class DbtProfileError(ErrorLevel, Cli, File):
    code: str = "A011"

    def message(self) -> str:
        return "Encountered an error while reading profiles:"


@dataclass
class DbtProfileErrorException(ErrorLevel, Cli, File):
    exc: Exception
    code: str = "A012"

    def message(self) -> str:
        return f"  ERROR: {str(self.exc)}"


@dataclass
class ProfileListTitle(InfoLevel, Cli, File):
    code: str = "A013"

    def message(self) -> str:
        return "Defined profiles:"


@dataclass
class ListSingleProfile(InfoLevel, Cli, File):
    profile: str
    code: str = "A014"

    def message(self) -> str:
        return f" - {self.profile}"


@dataclass
class NoDefinedProfiles(InfoLevel, Cli, File):
    code: str = "A015"

    def message(self) -> str:
        return "There are no profiles defined in your profiles.yml file"


@dataclass
class ProfileHelpMessage(InfoLevel, Cli, File):
    code: str = "A016"

    def message(self) -> str:
        PROFILES_HELP_MESSAGE = """
For more information on configuring profiles, please consult the dbt docs:

https://docs.getdbt.com/docs/configure-your-profile
"""
        return PROFILES_HELP_MESSAGE


@dataclass
class CatchableExceptionOnRun(ShowException, DebugLevel, Cli, File):
    exc: Exception
    code: str = "W002"

    def message(self) -> str:
        return str(self.exc)


@dataclass
class InternalExceptionOnRun(DebugLevel, Cli, File):
    build_path: str
    exc: Exception
    code: str = "W003"

    def message(self) -> str:
        prefix = 'Internal error executing {}'.format(self.build_path)

        INTERNAL_ERROR_STRING = """This is an error in dbt. Please try again. If \
the error persists, open an issue at https://github.com/dbt-labs/dbt-core
""".strip()

        return "{prefix}\n{error}\n\n{note}".format(
            prefix=ui.red(prefix),
            error=str(self.exc).strip(),
            note=INTERNAL_ERROR_STRING
        )


# This prints the stack trace at the debug level while allowing just the nice exception message
# at the error level - or whatever other level chosen.  Used in multiple places.
@dataclass
class PrintDebugStackTrace(ShowException, DebugLevel, Cli, File):
    code: str = "Z011"

    def message(self) -> str:
        return ""


@dataclass
class GenericExceptionOnRun(ErrorLevel, Cli, File):
    build_path: Optional[str]
    unique_id: str
    exc: str  # TODO: make this the actual exception once we have a better searilization strategy
    code: str = "W004"

    def message(self) -> str:
        node_description = self.build_path
        if node_description is None:
            node_description = self.unique_id
        prefix = "Unhandled error while executing {}".format(node_description)
        return "{prefix}\n{error}".format(
            prefix=ui.red(prefix),
            error=str(self.exc).strip()
        )


@dataclass
class NodeConnectionReleaseError(ShowException, DebugLevel, Cli, File):
    node_name: str
    exc: Exception
    code: str = "W005"

    def message(self) -> str:
        return ('Error releasing connection for node {}: {!s}'
                .format(self.node_name, self.exc))


@dataclass
class CheckCleanPath(InfoLevel, Cli):
    path: str
    code: str = "Z012"

    def message(self) -> str:
        return f"Checking {self.path}/*"


@dataclass
class ConfirmCleanPath(InfoLevel, Cli):
    path: str

    code: str = "Z013"

    def message(self) -> str:
        return f"Cleaned {self.path}/*"


@dataclass
class ProtectedCleanPath(InfoLevel, Cli):
    path: str
    code: str = "Z014"

    def message(self) -> str:
        return f"ERROR: not cleaning {self.path}/* because it is protected"


@dataclass
class FinishedCleanPaths(InfoLevel, Cli):
    code: str = "Z015"

    def message(self) -> str:
        return "Finished cleaning all paths."


@dataclass
class OpenCommand(InfoLevel, Cli, File):
    open_cmd: str
    profiles_dir: str
    code: str = "Z016"

    def message(self) -> str:
        PROFILE_DIR_MESSAGE = """To view your profiles.yml file, run:

{open_cmd} {profiles_dir}"""
        message = PROFILE_DIR_MESSAGE.format(
            open_cmd=self.open_cmd,
            profiles_dir=self.profiles_dir
        )

        return message


@dataclass
class DepsNoPackagesFound(InfoLevel, Cli, File):
    code: str = "M013"

    def message(self) -> str:
        return 'Warning: No packages were found in packages.yml'


@dataclass
class DepsStartPackageInstall(InfoLevel, Cli, File):
    package_name: str
    code: str = "M014"

    def message(self) -> str:
        return f"Installing {self.package_name}"


@dataclass
class DepsInstallInfo(InfoLevel, Cli, File):
    version_name: str
    code: str = "M015"

    def message(self) -> str:
        return f"  Installed from {self.version_name}"


@dataclass
class DepsUpdateAvailable(InfoLevel, Cli, File):
    version_latest: str
    code: str = "M016"

    def message(self) -> str:
        return f"  Updated version available: {self.version_latest}"


@dataclass
class DepsUTD(InfoLevel, Cli, File):
    code: str = "M017"

    def message(self) -> str:
        return "  Up to date!"


@dataclass
class DepsListSubdirectory(InfoLevel, Cli, File):
    subdirectory: str
    code: str = "M018"

    def message(self) -> str:
        return f"   and subdirectory {self.subdirectory}"


@dataclass
class DepsNotifyUpdatesAvailable(InfoLevel, Cli, File):
    packages: List[str]
    code: str = "M019"

    def message(self) -> str:
        return ('Updates available for packages: {} \
                \nUpdate your versions in packages.yml, then run dbt deps'.format(self.packages))


@dataclass
class DatabaseErrorRunning(InfoLevel, Cli, File):
    hook_type: str
    code: str = "E038"

    def message(self) -> str:
        return f"Database error while running {self.hook_type}"


@dataclass
class EmptyLine(InfoLevel, Cli, File):
    code: str = "Z017"

    def message(self) -> str:
        return ''


@dataclass
class HooksRunning(InfoLevel, Cli, File):
    num_hooks: int
    hook_type: str
    code: str = "E039"

    def message(self) -> str:
        plural = 'hook' if self.num_hooks == 1 else 'hooks'
        return f"Running {self.num_hooks} {self.hook_type} {plural}"


@dataclass
class HookFinished(InfoLevel, Cli, File):
    stat_line: str
    execution: str
    code: str = "E040"

    def message(self) -> str:
        return f"Finished running {self.stat_line}{self.execution}."


@dataclass
class WriteCatalogFailure(ErrorLevel, Cli, File):
    num_exceptions: int
    code: str = "E041"

    def message(self) -> str:
        return (f"dbt encountered {self.num_exceptions} failure{(self.num_exceptions != 1) * 's'} "
                "while writing the catalog")


@dataclass
class CatalogWritten(InfoLevel, Cli, File):
    path: str
    code: str = "E042"

    def message(self) -> str:
        return f"Catalog written to {self.path}"


@dataclass
class CannotGenerateDocs(InfoLevel, Cli, File):
    code: str = "E043"

    def message(self) -> str:
        return "compile failed, cannot generate docs"


@dataclass
class BuildingCatalog(InfoLevel, Cli, File):
    code: str = "E044"

    def message(self) -> str:
        return "Building catalog"


@dataclass
class CompileComplete(InfoLevel, Cli, File):
    code: str = "Q002"

    def message(self) -> str:
        return "Done."


@dataclass
class FreshnessCheckComplete(InfoLevel, Cli, File):
    code: str = "Q003"

    def message(self) -> str:
        return "Done."


@dataclass
class ServingDocsPort(InfoLevel, Cli, File):
    address: str
    port: int
    code: str = "Z018"

    def message(self) -> str:
        return f"Serving docs at {self.address}:{self.port}"


@dataclass
class ServingDocsAccessInfo(InfoLevel, Cli, File):
    port: str
    code: str = "Z019"

    def message(self) -> str:
        return f"To access from your browser, navigate to:  http://localhost:{self.port}"


@dataclass
class ServingDocsExitInfo(InfoLevel, Cli, File):
    code: str = "Z020"

    def message(self) -> str:
        return "Press Ctrl+C to exit."


@dataclass
class SeedHeader(InfoLevel, Cli, File):
    header: str
    code: str = "Q004"

    def message(self) -> str:
        return self.header


@dataclass
class SeedHeaderSeperator(InfoLevel, Cli, File):
    len_header: int
    code: str = "Q005"

    def message(self) -> str:
        return "-" * self.len_header


@dataclass
class RunResultWarning(WarnLevel, Cli, File):
    resource_type: str
    node_name: str
    path: str
    code: str = "Z021"

    def message(self) -> str:
        info = 'Warning'
        return ui.yellow(f"{info} in {self.resource_type} {self.node_name} ({self.path})")


@dataclass
class RunResultFailure(ErrorLevel, Cli, File):
    resource_type: str
    node_name: str
    path: str
    code: str = "Z022"

    def message(self) -> str:
        info = 'Failure'
        return ui.red(f"{info} in {self.resource_type} {self.node_name} ({self.path})")


@dataclass
class StatsLine(InfoLevel, Cli, File):
    stats: Dict
    code: str = "Z023"

    def message(self) -> str:
        stats_line = ("Done. PASS={pass} WARN={warn} ERROR={error} SKIP={skip} TOTAL={total}")
        return stats_line.format(**self.stats)


@dataclass
class RunResultError(ErrorLevel, Cli, File):
    msg: str
    code: str = "Z024"

    def message(self) -> str:
        return f"  {self.msg}"


@dataclass
class RunResultErrorNoMessage(ErrorLevel, Cli, File):
    status: str
    code: str = "Z025"

    def message(self) -> str:
        return f"  Status: {self.status}"


@dataclass
class SQLCompiledPath(InfoLevel, Cli, File):
    path: str
    code: str = "Z026"

    def message(self) -> str:
        return f"  compiled SQL at {self.path}"


@dataclass
class SQlRunnerException(ShowException, DebugLevel, Cli, File):
    exc: Exception
    code: str = "Q006"

    def message(self) -> str:
        return f"Got an exception: {self.exc}"


@dataclass
class CheckNodeTestFailure(InfoLevel, Cli, File):
    relation_name: str
    code: str = "Z027"

    def message(self) -> str:
        msg = f"select * from {self.relation_name}"
        border = '-' * len(msg)
        return f"  See test failures:\n  {border}\n  {msg}\n  {border}"


@dataclass
class FirstRunResultError(ErrorLevel, Cli, File):
    msg: str
    code: str = "Z028"

    def message(self) -> str:
        return ui.yellow(self.msg)


@dataclass
class AfterFirstRunResultError(ErrorLevel, Cli, File):
    msg: str
    code: str = "Z029"

    def message(self) -> str:
        return self.msg


@dataclass
class EndOfRunSummary(InfoLevel, Cli, File):
    num_errors: int
    num_warnings: int
    keyboard_interrupt: bool = False
    code: str = "Z030"

    def message(self) -> str:
        error_plural = pluralize(self.num_errors, 'error')
        warn_plural = pluralize(self.num_warnings, 'warning')
        if self.keyboard_interrupt:
            message = ui.yellow('Exited because of keyboard interrupt.')
        elif self.num_errors > 0:
            message = ui.red("Completed with {} and {}:".format(
                error_plural, warn_plural))
        elif self.num_warnings > 0:
            message = ui.yellow('Completed with {}:'.format(warn_plural))
        else:
            message = ui.green('Completed successfully')
        return message


@dataclass
class PrintStartLine(InfoLevel, Cli, File, NodeInfo):
    description: str
    index: int
    total: int
    report_node_data: ParsedModelNode
    code: str = "Q033"

    def message(self) -> str:
        msg = f"START {self.description}"
        return format_fancy_output_line(
            msg=msg,
            status='RUN',
            index=self.index,
            total=self.total
        )


@dataclass
class PrintHookStartLine(InfoLevel, Cli, File, NodeInfo):
    statement: str
    index: int
    total: int
    truncate: bool
    report_node_data: Any  # TODO: resolve ParsedHookNode circular import
    code: str = "Q032"

    def message(self) -> str:
        msg = f"START hook: {self.statement}"
        return format_fancy_output_line(msg=msg,
                                        status='RUN',
                                        index=self.index,
                                        total=self.total,
                                        truncate=self.truncate)


@dataclass
class PrintHookEndLine(InfoLevel, Cli, File, NodeInfo):
    statement: str
    status: str
    index: int
    total: int
    execution_time: int
    truncate: bool
    report_node_data: Any  # TODO: resolve ParsedHookNode circular import
    code: str = "Q007"

    def message(self) -> str:
        msg = 'OK hook: {}'.format(self.statement)
        return format_fancy_output_line(msg=msg,
                                        status=ui.green(self.status),
                                        index=self.index,
                                        total=self.total,
                                        execution_time=self.execution_time,
                                        truncate=self.truncate)


@dataclass
class SkippingDetails(InfoLevel, Cli, File, NodeInfo):
    resource_type: str
    schema: str
    node_name: str
    index: int
    total: int
    report_node_data: ParsedModelNode
    code: str = "Q034"

    def message(self) -> str:
        if self.resource_type in NodeType.refable():
            msg = f'SKIP relation {self.schema}.{self.node_name}'
        else:
            msg = f'SKIP {self.resource_type} {self.node_name}'
        return format_fancy_output_line(msg=msg,
                                        status=ui.yellow('SKIP'),
                                        index=self.index,
                                        total=self.total)


@dataclass
class PrintErrorTestResult(ErrorLevel, Cli, File, NodeInfo):
    name: str
    index: int
    num_models: int
    execution_time: int
    report_node_data: ParsedModelNode
    code: str = "Q008"

    def message(self) -> str:
        info = "ERROR"
        msg = f"{info} {self.name}"
        return format_fancy_output_line(msg=msg,
                                        status=ui.red(info),
                                        index=self.index,
                                        total=self.num_models,
                                        execution_time=self.execution_time)


@dataclass
class PrintPassTestResult(InfoLevel, Cli, File, NodeInfo):
    name: str
    index: int
    num_models: int
    execution_time: int
    report_node_data: ParsedModelNode
    code: str = "Q009"

    def message(self) -> str:
        info = "PASS"
        msg = f"{info} {self.name}"
        return format_fancy_output_line(msg=msg,
                                        status=ui.green(info),
                                        index=self.index,
                                        total=self.num_models,
                                        execution_time=self.execution_time)


@dataclass
class PrintWarnTestResult(WarnLevel, Cli, File, NodeInfo):
    name: str
    index: int
    num_models: int
    execution_time: int
    failures: List[str]
    report_node_data: ParsedModelNode
    code: str = "Q010"

    def message(self) -> str:
        info = f'WARN {self.failures}'
        msg = f"{info} {self.name}"
        return format_fancy_output_line(msg=msg,
                                        status=ui.yellow(info),
                                        index=self.index,
                                        total=self.num_models,
                                        execution_time=self.execution_time)


@dataclass
class PrintFailureTestResult(ErrorLevel, Cli, File, NodeInfo):
    name: str
    index: int
    num_models: int
    execution_time: int
    failures: List[str]
    report_node_data: ParsedModelNode
    code: str = "Q011"

    def message(self) -> str:
        info = f'FAIL {self.failures}'
        msg = f"{info} {self.name}"
        return format_fancy_output_line(msg=msg,
                                        status=ui.red(info),
                                        index=self.index,
                                        total=self.num_models,
                                        execution_time=self.execution_time)


@dataclass
class PrintSkipBecauseError(ErrorLevel, Cli, File):
    schema: str
    relation: str
    index: int
    total: int
    code: str = "Z034"

    def message(self) -> str:
        msg = f'SKIP relation {self.schema}.{self.relation} due to ephemeral model error'
        return format_fancy_output_line(msg=msg,
                                        status=ui.red('ERROR SKIP'),
                                        index=self.index,
                                        total=self.total)


@dataclass
class PrintModelErrorResultLine(ErrorLevel, Cli, File, NodeInfo):
    description: str
    status: str
    index: int
    total: int
    execution_time: int
    report_node_data: ParsedModelNode
    code: str = "Q035"

    def message(self) -> str:
        info = "ERROR creating"
        msg = f"{info} {self.description}"
        return format_fancy_output_line(msg=msg,
                                        status=ui.red(self.status.upper()),
                                        index=self.index,
                                        total=self.total,
                                        execution_time=self.execution_time)


@dataclass
class PrintModelResultLine(InfoLevel, Cli, File, NodeInfo):
    description: str
    status: str
    index: int
    total: int
    execution_time: int
    report_node_data: ParsedModelNode
    code: str = "Q012"

    def message(self) -> str:
        info = "OK created"
        msg = f"{info} {self.description}"
        return format_fancy_output_line(msg=msg,
                                        status=ui.green(self.status),
                                        index=self.index,
                                        total=self.total,
                                        execution_time=self.execution_time)


@dataclass
class PrintSnapshotErrorResultLine(ErrorLevel, Cli, File, NodeInfo):
    status: str
    description: str
    cfg: Dict
    index: int
    total: int
    execution_time: int
    report_node_data: ParsedModelNode
    code: str = "Q013"

    def message(self) -> str:
        info = 'ERROR snapshotting'
        msg = "{info} {description}".format(info=info, description=self.description, **self.cfg)
        return format_fancy_output_line(msg=msg,
                                        status=ui.red(self.status.upper()),
                                        index=self.index,
                                        total=self.total,
                                        execution_time=self.execution_time)


@dataclass
class PrintSnapshotResultLine(InfoLevel, Cli, File, NodeInfo):
    status: str
    description: str
    cfg: Dict
    index: int
    total: int
    execution_time: int
    report_node_data: ParsedModelNode
    code: str = "Q014"

    def message(self) -> str:
        info = 'OK snapshotted'
        msg = "{info} {description}".format(info=info, description=self.description, **self.cfg)
        return format_fancy_output_line(msg=msg,
                                        status=ui.green(self.status),
                                        index=self.index,
                                        total=self.total,
                                        execution_time=self.execution_time)


@dataclass
class PrintSeedErrorResultLine(ErrorLevel, Cli, File, NodeInfo):
    status: str
    index: int
    total: int
    execution_time: int
    schema: str
    relation: str
    report_node_data: ParsedModelNode
    code: str = "Q015"

    def message(self) -> str:
        info = 'ERROR loading'
        msg = f"{info} seed file {self.schema}.{self.relation}"
        return format_fancy_output_line(msg=msg,
                                        status=ui.red(self.status.upper()),
                                        index=self.index,
                                        total=self.total,
                                        execution_time=self.execution_time)


@dataclass
class PrintSeedResultLine(InfoLevel, Cli, File, NodeInfo):
    status: str
    index: int
    total: int
    execution_time: int
    schema: str
    relation: str
    report_node_data: ParsedModelNode
    code: str = "Q016"

    def message(self) -> str:
        info = 'OK loaded'
        msg = f"{info} seed file {self.schema}.{self.relation}"
        return format_fancy_output_line(msg=msg,
                                        status=ui.green(self.status),
                                        index=self.index,
                                        total=self.total,
                                        execution_time=self.execution_time)


@dataclass
class PrintHookEndErrorLine(ErrorLevel, Cli, File, NodeInfo):
    source_name: str
    table_name: str
    index: int
    total: int
    execution_time: int
    report_node_data: ParsedHookNode
    code: str = "Q017"

    def message(self) -> str:
        info = 'ERROR'
        msg = f"{info} freshness of {self.source_name}.{self.table_name}"
        return format_fancy_output_line(msg=msg,
                                        status=ui.red(info),
                                        index=self.index,
                                        total=self.total,
                                        execution_time=self.execution_time)


@dataclass
class PrintHookEndErrorStaleLine(ErrorLevel, Cli, File, NodeInfo):
    source_name: str
    table_name: str
    index: int
    total: int
    execution_time: int
    report_node_data: ParsedHookNode
    code: str = "Q018"

    def message(self) -> str:
        info = 'ERROR STALE'
        msg = f"{info} freshness of {self.source_name}.{self.table_name}"
        return format_fancy_output_line(msg=msg,
                                        status=ui.red(info),
                                        index=self.index,
                                        total=self.total,
                                        execution_time=self.execution_time)


@dataclass
class PrintHookEndWarnLine(WarnLevel, Cli, File, NodeInfo):
    source_name: str
    table_name: str
    index: int
    total: int
    execution_time: int
    report_node_data: ParsedHookNode
    code: str = "Q019"

    def message(self) -> str:
        info = 'WARN'
        msg = f"{info} freshness of {self.source_name}.{self.table_name}"
        return format_fancy_output_line(msg=msg,
                                        status=ui.yellow(info),
                                        index=self.index,
                                        total=self.total,
                                        execution_time=self.execution_time)


@dataclass
class PrintHookEndPassLine(InfoLevel, Cli, File, NodeInfo):
    source_name: str
    table_name: str
    index: int
    total: int
    execution_time: int
    report_node_data: ParsedHookNode
    code: str = "Q020"

    def message(self) -> str:
        info = 'PASS'
        msg = f"{info} freshness of {self.source_name}.{self.table_name}"
        return format_fancy_output_line(msg=msg,
                                        status=ui.green(info),
                                        index=self.index,
                                        total=self.total,
                                        execution_time=self.execution_time)


@dataclass
class PrintCancelLine(ErrorLevel, Cli, File):
    conn_name: str
    code: str = "Q021"

    def message(self) -> str:
        msg = 'CANCEL query {}'.format(self.conn_name)
        return format_fancy_output_line(msg=msg,
                                        status=ui.red('CANCEL'),
                                        index=None,
                                        total=None)


@dataclass
class DefaultSelector(InfoLevel, Cli, File):
    name: str
    code: str = "Q022"

    def message(self) -> str:
        return f"Using default selector {self.name}"


@dataclass
class NodeStart(DebugLevel, Cli, File, NodeInfo):
    unique_id: str
    report_node_data: ParsedModelNode
    code: str = "Q023"

    def message(self) -> str:
        return f"Began running node {self.unique_id}"


@dataclass
class NodeFinished(DebugLevel, Cli, File, NodeInfo):
    unique_id: str
    report_node_data: ParsedModelNode
    run_result: RunResult
    code: str = "Q024"

    def message(self) -> str:
        return f"Finished running node {self.unique_id}"

    @classmethod
    def asdict(cls, data: list) -> dict:
        return dict((k, str(v)) for k, v in data)


@dataclass
class QueryCancelationUnsupported(InfoLevel, Cli, File):
    type: str
    code: str = "Q025"

    def message(self) -> str:
        msg = (f"The {self.type} adapter does not support query "
               "cancellation. Some queries may still be "
               "running!")
        return ui.yellow(msg)


@dataclass
class ConcurrencyLine(InfoLevel, Cli, File):
    num_threads: int
    target_name: str
    code: str = "Q026"

    def message(self) -> str:
        return f"Concurrency: {self.num_threads} threads (target='{self.target_name}')"


@dataclass
class NodeCompiling(DebugLevel, Cli, File, NodeInfo):
    unique_id: str
    report_node_data: ParsedModelNode
    code: str = "Q030"

    def message(self) -> str:
        return f"Began compiling node {self.unique_id}"


@dataclass
class NodeExecuting(DebugLevel, Cli, File, NodeInfo):
    unique_id: str
    report_node_data: ParsedModelNode
    code: str = "Q031"

    def message(self) -> str:
        return f"Began executing node {self.unique_id}"


@dataclass
class StarterProjectPath(DebugLevel, Cli, File):
    dir: str
    code: str = "A017"

    def message(self) -> str:
        return f"Starter project path: {self.dir}"


@dataclass
class ConfigFolderDirectory(InfoLevel, Cli, File):
    dir: str
    code: str = "A018"

    def message(self) -> str:
        return f"Creating dbt configuration folder at {self.dir}"


@dataclass
class NoSampleProfileFound(InfoLevel, Cli, File):
    adapter: str
    code: str = "A019"

    def message(self) -> str:
        return f"No sample profile found for {self.adapter}."


@dataclass
class ProfileWrittenWithSample(InfoLevel, Cli, File):
    name: str
    path: str
    code: str = "A020"

    def message(self) -> str:
        return (f"Profile {self.name} written to {self.path} "
                "using target's sample configuration. Once updated, you'll be able to "
                "start developing with dbt.")


@dataclass
class ProfileWrittenWithTargetTemplateYAML(InfoLevel, Cli, File):
    name: str
    path: str
    code: str = "A021"

    def message(self) -> str:
        return (f"Profile {self.name} written to {self.path} using target's "
                "profile_template.yml and your supplied values. Run 'dbt debug' to "
                "validate the connection.")


@dataclass
class ProfileWrittenWithProjectTemplateYAML(InfoLevel, Cli, File):
    name: str
    path: str
    code: str = "A022"

    def message(self) -> str:
        return (f"Profile {self.name} written to {self.path} using project's "
                "profile_template.yml and your supplied values. Run 'dbt debug' to "
                "validate the connection.")


@dataclass
class SettingUpProfile(InfoLevel, Cli, File):
    code: str = "A023"

    def message(self) -> str:
        return "Setting up your profile."


@dataclass
class InvalidProfileTemplateYAML(InfoLevel, Cli, File):
    code: str = "A024"

    def message(self) -> str:
        return "Invalid profile_template.yml in project."


@dataclass
class ProjectNameAlreadyExists(InfoLevel, Cli, File):
    name: str
    code: str = "A025"

    def message(self) -> str:
        return f"A project called {self.name} already exists here."


@dataclass
class GetAddendum(InfoLevel, Cli, File):
    msg: str
    code: str = "A026"

    def message(self) -> str:
        return self.msg


@dataclass
class DepsSetDownloadDirectory(DebugLevel, Cli, File):
    path: str
    code: str = "A027"

    def message(self) -> str:
        return f"Set downloads directory='{self.path}'"


@dataclass
class EnsureGitInstalled(ErrorLevel, Cli, File):
    code: str = "Z036"

    def message(self) -> str:
        return ('Make sure git is installed on your machine. More '
                'information: '
                'https://docs.getdbt.com/docs/package-management')


@dataclass
class DepsCreatingLocalSymlink(DebugLevel, Cli, File):
    code: str = "Z037"

    def message(self) -> str:
        return '  Creating symlink to local dependency.'


@dataclass
class DepsSymlinkNotAvailable(DebugLevel, Cli, File):
    code: str = "Z038"

    def message(self) -> str:
        return '  Symlinks are not available on this OS, copying dependency.'


@dataclass
class FoundStats(InfoLevel, Cli, File):
    stat_line: str
    code: str = "W006"

    def message(self) -> str:
        return f"Found {self.stat_line}"


# TODO: should this have NodeInfo on it?
@dataclass
class CompilingNode(DebugLevel, Cli, File):
    unique_id: str
    code: str = "Q027"

    def message(self) -> str:
        return f"Compiling {self.unique_id}"


@dataclass
class WritingInjectedSQLForNode(DebugLevel, Cli, File):
    unique_id: str
    code: str = "Q028"

    def message(self) -> str:
        return f'Writing injected SQL for node "{self.unique_id}"'


@dataclass
class DisableTracking(WarnLevel, Cli, File):
    code: str = "Z039"

    def message(self) -> str:
        return "Error sending message, disabling tracking"


@dataclass
class SendingEvent(DebugLevel, Cli, File):
    kwargs: str
    code: str = "Z040"

    def message(self) -> str:
        return f"Sending event: {self.kwargs}"


@dataclass
class SendEventFailure(DebugLevel, Cli, File):
    code: str = "Z041"

    def message(self) -> str:
        return "An error was encountered while trying to send an event"


@dataclass
class FlushEvents(DebugLevel, Cli):
    code: str = "Z042"

    def message(self) -> str:
        return "Flushing usage events"


@dataclass
class FlushEventsFailure(DebugLevel, Cli):
    code: str = "Z043"

    def message(self) -> str:
        return "An error was encountered while trying to flush usage events"


@dataclass
class TrackingInitializeFailure(ShowException, DebugLevel, Cli, File):
    code: str = "Z044"

    def message(self) -> str:
        return "Got an exception trying to initialize tracking"


@dataclass
class RetryExternalCall(DebugLevel, Cli, File):
    attempt: int
    max: int
    code: str = "Z045"

    def message(self) -> str:
        return f"Retrying external call. Attempt: {self.attempt} Max attempts: {self.max}"


@dataclass
class GeneralWarningMsg(WarnLevel, Cli, File):
    msg: str
    log_fmt: str
    code: str = "Z046"

    def message(self) -> str:
        if self.log_fmt is not None:
            return self.log_fmt.format(self.msg)
        return self.msg


@dataclass
class GeneralWarningException(WarnLevel, Cli, File):
    exc: Exception
    log_fmt: str
    code: str = "Z047"

    def message(self) -> str:
        if self.log_fmt is not None:
            return self.log_fmt.format(str(self.exc))
        return str(self.exc)


@dataclass
class EventBufferFull(WarnLevel, Cli, File):
    code: str = "Z048"

    def message(self) -> str:
        return "Internal event buffer full. Earliest events will be dropped (FIFO)."


# since mypy doesn't run on every file we need to suggest to mypy that every
# class gets instantiated. But we don't actually want to run this code.
# making the conditional `if False` causes mypy to skip it as dead code so
# we need to skirt around that by computing something it doesn't check statically.
#
# TODO remove these lines once we run mypy everywhere.
if 1 == 0:
    MainReportVersion('')
    MainKeyboardInterrupt()
    MainEncounteredError(BaseException(''))
    MainStackTrace('')
    MainTrackingUserState('')
    ParsingStart()
    ParsingCompiling()
    ParsingWritingManifest()
    ParsingDone()
    ManifestDependenciesLoaded()
    ManifestLoaderCreated()
    ManifestLoaded()
    ManifestChecked()
    ManifestFlatGraphBuilt()
    ReportPerformancePath(path="")
    GitSparseCheckoutSubdirectory(subdir="")
    GitProgressCheckoutRevision(revision="")
    GitProgressUpdatingExistingDependency(dir="")
    GitProgressPullingNewDependency(dir="")
    GitNothingToDo(sha="")
    GitProgressUpdatedCheckoutRange(start_sha="", end_sha="")
    GitProgressCheckedOutAt(end_sha="")
    SystemErrorRetrievingModTime(path="")
    SystemCouldNotWrite(path="", reason="", exc=Exception(""))
    SystemExecutingCmd(cmd=[""])
    SystemStdOutMsg(bmsg=b"")
    SystemStdErrMsg(bmsg=b"")
    SelectorReportInvalidSelector(
        selector_methods={"": ""}, spec_method="", raw_spec=""
    )
    MacroEventInfo(msg="")
    MacroEventDebug(msg="")
    NewConnection(conn_type="", conn_name="")
    ConnectionReused(conn_name="")
    ConnectionLeftOpen(conn_name="")
    ConnectionClosed(conn_name="")
    RollbackFailed(conn_name="")
    ConnectionClosed2(conn_name="")
    ConnectionLeftOpen2(conn_name="")
    Rollback(conn_name="")
    CacheMiss(conn_name="", database="", schema="")
    ListRelations(database="", schema="", relations=[])
    ConnectionUsed(conn_type="", conn_name="")
    SQLQuery(conn_name="", sql="")
    SQLQueryStatus(status="", elapsed=0.1)
    SQLCommit(conn_name="")
    ColTypeChange(orig_type="", new_type="", table="")
    SchemaCreation(relation=_make_key(BaseRelation()))
    SchemaDrop(relation=_make_key(BaseRelation()))
    UncachedRelation(
        dep_key=_ReferenceKey(database="", schema="", identifier=""),
        ref_key=_ReferenceKey(database="", schema="", identifier=""),
    )
    AddLink(
        dep_key=_ReferenceKey(database="", schema="", identifier=""),
        ref_key=_ReferenceKey(database="", schema="", identifier=""),
    )
    AddRelation(relation=_make_key(_CachedRelation()))
    DropMissingRelation(relation=_ReferenceKey(database="", schema="", identifier=""))
    DropCascade(
        dropped=_ReferenceKey(database="", schema="", identifier=""),
        consequences={_ReferenceKey(database="", schema="", identifier="")},
    )
    UpdateReference(
        old_key=_ReferenceKey(database="", schema="", identifier=""),
        new_key=_ReferenceKey(database="", schema="", identifier=""),
        cached_key=_ReferenceKey(database="", schema="", identifier=""),
    )
    TemporaryRelation(key=_ReferenceKey(database="", schema="", identifier=""))
    RenameSchema(
        old_key=_ReferenceKey(database="", schema="", identifier=""),
        new_key=_ReferenceKey(database="", schema="", identifier="")
    )
    DumpBeforeAddGraph(dict())
    DumpAfterAddGraph(dict())
    DumpBeforeRenameSchema(dict())
    DumpAfterRenameSchema(dict())
    AdapterImportError(ModuleNotFoundError())
    PluginLoadError()
    SystemReportReturnCode(returncode=0)
    NewConnectionOpening(connection_state='')
    TimingInfoCollected()
    MergedFromState(nbr_merged=0, sample=[])
    MissingProfileTarget(profile_name='', target_name='')
    InvalidVarsYAML()
    GenericTestFileParse(path='')
    MacroFileParse(path='')
    PartialParsingFullReparseBecauseOfError()
    PartialParsingFile(file_dict={})
    PartialParsingExceptionFile(file='')
    PartialParsingException(exc_info={})
    PartialParsingSkipParsing()
    PartialParsingMacroChangeStartFullParse()
    ManifestWrongMetadataVersion(version='')
    PartialParsingVersionMismatch(saved_version='', current_version='')
    PartialParsingFailedBecauseConfigChange()
    PartialParsingFailedBecauseProfileChange()
    PartialParsingFailedBecauseNewProjectDependency()
    PartialParsingFailedBecauseHashChanged()
    PartialParsingDeletedMetric('')
    ParsedFileLoadFailed(path='', exc=Exception(''))
    PartialParseSaveFileNotFound()
    StaticParserCausedJinjaRendering(path='')
    UsingExperimentalParser(path='')
    SampleFullJinjaRendering(path='')
    StaticParserFallbackJinjaRendering(path='')
    StaticParsingMacroOverrideDetected(path='')
    StaticParserSuccess(path='')
    StaticParserFailure(path='')
    ExperimentalParserSuccess(path='')
    ExperimentalParserFailure(path='')
    PartialParsingEnabled(deleted=0, added=0, changed=0)
    PartialParsingAddedFile(file_id='')
    PartialParsingDeletedFile(file_id='')
    PartialParsingUpdatedFile(file_id='')
    PartialParsingNodeMissingInSourceFile(source_file='')
    PartialParsingMissingNodes(file_id='')
    PartialParsingChildMapMissingUniqueID(unique_id='')
    PartialParsingUpdateSchemaFile(file_id='')
    PartialParsingDeletedSource(unique_id='')
    PartialParsingDeletedExposure(unique_id='')
    InvalidDisabledSourceInTestNode(msg='')
    InvalidRefInTestNode(msg='')
    RunningOperationCaughtError(exc=Exception(''))
    RunningOperationUncaughtError(exc=Exception(''))
    DbtProjectError()
    DbtProjectErrorException(exc=Exception(''))
    DbtProfileError()
    DbtProfileErrorException(exc=Exception(''))
    ProfileListTitle()
    ListSingleProfile(profile='')
    NoDefinedProfiles()
    ProfileHelpMessage()
    CatchableExceptionOnRun(exc=Exception(''))
    InternalExceptionOnRun(build_path='', exc=Exception(''))
    GenericExceptionOnRun(build_path='', unique_id='', exc='')
    NodeConnectionReleaseError(node_name='', exc=Exception(''))
    CheckCleanPath(path='')
    ConfirmCleanPath(path='')
    ProtectedCleanPath(path='')
    FinishedCleanPaths()
    OpenCommand(open_cmd='', profiles_dir='')
    DepsNoPackagesFound()
    DepsStartPackageInstall(package_name='')
    DepsInstallInfo(version_name='')
    DepsUpdateAvailable(version_latest='')
    DepsListSubdirectory(subdirectory='')
    DepsNotifyUpdatesAvailable(packages=[])
    DatabaseErrorRunning(hook_type='')
    EmptyLine()
    HooksRunning(num_hooks=0, hook_type='')
    HookFinished(stat_line='', execution='')
    WriteCatalogFailure(num_exceptions=0)
    CatalogWritten(path='')
    CannotGenerateDocs()
    BuildingCatalog()
    CompileComplete()
    FreshnessCheckComplete()
    ServingDocsPort(address='', port=0)
    ServingDocsAccessInfo(port='')
    ServingDocsExitInfo()
    SeedHeader(header='')
    SeedHeaderSeperator(len_header=0)
    RunResultWarning(resource_type='', node_name='', path='')
    RunResultFailure(resource_type='', node_name='', path='')
    StatsLine(stats={})
    RunResultError(msg='')
    RunResultErrorNoMessage(status='')
    SQLCompiledPath(path='')
    CheckNodeTestFailure(relation_name='')
    FirstRunResultError(msg='')
    AfterFirstRunResultError(msg='')
    EndOfRunSummary(num_errors=0, num_warnings=0, keyboard_interrupt=False)
    PrintStartLine(description='', index=0, total=0, report_node_data=ParsedModelNode())
    PrintHookStartLine(
        statement='',
        index=0,
        total=0,
        truncate=False,
        report_node_data=ParsedHookNode()
    )
    PrintHookEndLine(
        statement='',
        status='',
        index=0,
        total=0,
        execution_time=0,
        truncate=False,
        report_node_data=ParsedHookNode()
    )
    SkippingDetails(
        resource_type='',
        schema='',
        node_name='',
        index=0,
        total=0,
        report_node_data=ParsedModelNode()
    )
    PrintErrorTestResult(
        name='',
        index=0,
        num_models=0,
        execution_time=0,
        report_node_data=ParsedModelNode()
    )
    PrintPassTestResult(
        name='',
        index=0,
        num_models=0,
        execution_time=0,
        report_node_data=ParsedModelNode()
    )
    PrintWarnTestResult(
        name='',
        index=0,
        num_models=0,
        execution_time=0,
        failures=[],
        report_node_data=ParsedModelNode()
    )
    PrintFailureTestResult(
        name='',
        index=0,
        num_models=0,
        execution_time=0,
        failures=[],
        report_node_data=ParsedModelNode()
    )
    PrintSkipBecauseError(schema='', relation='', index=0, total=0)
    PrintModelErrorResultLine(
        description='',
        status='',
        index=0,
        total=0,
        execution_time=0,
        report_node_data=ParsedModelNode()
    )
    PrintModelResultLine(
        description='',
        status='',
        index=0,
        total=0,
        execution_time=0,
        report_node_data=ParsedModelNode()
    )
    PrintSnapshotErrorResultLine(
        status='',
        description='',
        cfg={},
        index=0,
        total=0,
        execution_time=0,
        report_node_data=ParsedModelNode()
    )
    PrintSnapshotResultLine(
        status='',
        description='',
        cfg={},
        index=0,
        total=0,
        execution_time=0,
        report_node_data=ParsedModelNode()
    )
    PrintSeedErrorResultLine(
        status='',
        index=0,
        total=0,
        execution_time=0,
        schema='',
        relation='',
        report_node_data=ParsedModelNode()
    )
    PrintSeedResultLine(
        status='',
        index=0,
        total=0,
        execution_time=0,
        schema='',
        relation='',
        report_node_data=ParsedModelNode()
    )
    PrintHookEndErrorLine(
        source_name='',
        table_name='',
        index=0,
        total=0,
        execution_time=0,
        report_node_data=ParsedHookNode()
    )
    PrintHookEndErrorStaleLine(
        source_name='',
        table_name='',
        index=0,
        total=0,
        execution_time=0,
        report_node_data=ParsedHookNode()
    )
    PrintHookEndWarnLine(
        source_name='',
        table_name='',
        index=0,
        total=0,
        execution_time=0,
        report_node_data=ParsedHookNode()
    )
    PrintHookEndPassLine(
        source_name='',
        table_name='',
        index=0,
        total=0,
        execution_time=0,
        report_node_data=ParsedHookNode()
    )
    PrintCancelLine(conn_name='')
    DefaultSelector(name='')
    NodeStart(report_node_data=ParsedModelNode(), unique_id='')
    NodeFinished(report_node_data=ParsedModelNode(), unique_id='', run_result=RunResult())
    QueryCancelationUnsupported(type='')
    ConcurrencyLine(num_threads=0, target_name='')
    NodeCompiling(report_node_data=ParsedModelNode(), unique_id='')
    NodeExecuting(report_node_data=ParsedModelNode(), unique_id='')
    StarterProjectPath(dir='')
    ConfigFolderDirectory(dir='')
    NoSampleProfileFound(adapter='')
    ProfileWrittenWithSample(name='', path='')
    ProfileWrittenWithTargetTemplateYAML(name='', path='')
    ProfileWrittenWithProjectTemplateYAML(name='', path='')
    SettingUpProfile()
    InvalidProfileTemplateYAML()
    ProjectNameAlreadyExists(name='')
    GetAddendum(msg='')
    DepsSetDownloadDirectory(path='')
    EnsureGitInstalled()
    DepsCreatingLocalSymlink()
    DepsSymlinkNotAvailable()
    FoundStats(stat_line='')
    CompilingNode(unique_id='')
    WritingInjectedSQLForNode(unique_id='')
    DisableTracking()
    SendingEvent(kwargs='')
    SendEventFailure()
    FlushEvents()
    FlushEventsFailure()
    TrackingInitializeFailure()
    RetryExternalCall(attempt=0, max=0)
    GeneralWarningMsg(msg='', log_fmt='')
    GeneralWarningException(exc=Exception(''), log_fmt='')
    EventBufferFull()



from copy import deepcopy
from dbt.context.context_config import ContextConfig
from dbt.contracts.graph.parsed import ParsedModelNode
import dbt.flags as flags
from dbt.events.functions import fire_event
from dbt.events.types import (
    StaticParserCausedJinjaRendering, UsingExperimentalParser,
    SampleFullJinjaRendering, StaticParserFallbackJinjaRendering,
    StaticParsingMacroOverrideDetected, StaticParserSuccess, StaticParserFailure,
    ExperimentalParserSuccess, ExperimentalParserFailure
)
from dbt.node_types import NodeType
from dbt.parser.base import SimpleSQLParser
from dbt.parser.search import FileBlock
import dbt.tracking as tracking
from dbt import utils
from dbt_extractor import ExtractionError, py_extract_from_source  # type: ignore
from functools import reduce
from itertools import chain
import random
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union


class ModelParser(SimpleSQLParser[ParsedModelNode]):
    def parse_from_dict(self, dct, validate=True) -> ParsedModelNode:
        if validate:
            ParsedModelNode.validate(dct)
        return ParsedModelNode.from_dict(dct)

    @property
    def resource_type(self) -> NodeType:
        return NodeType.Model

    @classmethod
    def get_compiled_path(cls, block: FileBlock):
        return block.path.relative_path

    def render_update(
        self, node: ParsedModelNode, config: ContextConfig
    ) -> None:
        self.manifest._parsing_info.static_analysis_path_count += 1

        if not flags.STATIC_PARSER:
            # jinja rendering
            super().render_update(node, config)
            fire_event(StaticParserCausedJinjaRendering(path=node.path))
            return

        # only sample for experimental parser correctness on normal runs,
        # not when the experimental parser flag is on.
        exp_sample: bool = False
        # sampling the stable static parser against jinja is significantly
        # more expensive and therefore done far less frequently.
        stable_sample: bool = False
        # there are two samples above, and it is perfectly fine if both happen
        # at the same time. If that happens, the experimental parser, stable
        # parser, and jinja rendering will run on the same model file and
        # send back codes for experimental v stable, and stable v jinja.
        if not flags.USE_EXPERIMENTAL_PARSER:
            # `True` roughly 1/5000 times this function is called
            # sample = random.randint(1, 5001) == 5000
            stable_sample = random.randint(1, 5001) == 5000
            # sampling the experimental parser is explicitly disabled here, but use the following
            # commented code to sample a fraction of the time when new
            # experimental features are added.
            # `True` roughly 1/100 times this function is called
            # exp_sample = random.randint(1, 101) == 100

        # top-level declaration of variables
        statically_parsed: Optional[Union[str, Dict[str, List[Any]]]] = None
        experimental_sample: Optional[Union[str, Dict[str, List[Any]]]] = None
        exp_sample_node: Optional[ParsedModelNode] = None
        exp_sample_config: Optional[ContextConfig] = None
        jinja_sample_node: Optional[ParsedModelNode] = None
        jinja_sample_config: Optional[ContextConfig] = None
        result: List[str] = []

        # sample the experimental parser only during a normal run
        if exp_sample and not flags.USE_EXPERIMENTAL_PARSER:
            fire_event(UsingExperimentalParser(path=node.path))
            experimental_sample = self.run_experimental_parser(node)
            # if the experimental parser succeeded, make a full copy of model parser
            # and populate _everything_ into it so it can be compared apples-to-apples
            # with a fully jinja-rendered project. This is necessary because the experimental
            # parser will likely add features that the existing static parser will fail on
            # so comparing those directly would give us bad results. The comparison will be
            # conducted after this model has been fully rendered either by the static parser
            # or by full jinja rendering
            if isinstance(experimental_sample, dict):
                model_parser_copy = self.partial_deepcopy()
                exp_sample_node = deepcopy(node)
                exp_sample_config = deepcopy(config)
                model_parser_copy.populate(
                    exp_sample_node,
                    exp_sample_config,
                    experimental_sample
                )
        # use the experimental parser exclusively if the flag is on
        if flags.USE_EXPERIMENTAL_PARSER:
            statically_parsed = self.run_experimental_parser(node)
        # run the stable static parser unless it is explicitly turned off
        else:
            statically_parsed = self.run_static_parser(node)

        # if the static parser succeeded, extract some data in easy-to-compare formats
        if isinstance(statically_parsed, dict):
            # only sample jinja for the purpose of comparing with the stable static parser
            # if we know we don't need to fall back to jinja (i.e. - nothing to compare
            # with jinja v jinja).
            # This means we skip sampling for 40% of the 1/5000 samples. We could run the
            # sampling rng here, but the effect would be the same since we would only roll
            # it 40% of the time. So I've opted to keep all the rng code colocated above.
            if stable_sample and not flags.USE_EXPERIMENTAL_PARSER:
                fire_event(SampleFullJinjaRendering(path=node.path))
                # if this will _never_ mutate anything `self` we could avoid these deep copies,
                # but we can't really guarantee that going forward.
                model_parser_copy = self.partial_deepcopy()
                jinja_sample_node = deepcopy(node)
                jinja_sample_config = deepcopy(config)
                # rendering mutates the node and the config
                super(ModelParser, model_parser_copy) \
                    .render_update(jinja_sample_node, jinja_sample_config)

            # update the unrendered config with values from the static parser.
            # values from yaml files are in there already
            self.populate(
                node,
                config,
                statically_parsed
            )

            # if we took a jinja sample, compare now that the base node has been populated
            if jinja_sample_node is not None and jinja_sample_config is not None:
                result = _get_stable_sample_result(
                    jinja_sample_node,
                    jinja_sample_config,
                    node,
                    config
                )

            # if we took an experimental sample, compare now that the base node has been populated
            if exp_sample_node is not None and exp_sample_config is not None:
                result = _get_exp_sample_result(
                    exp_sample_node,
                    exp_sample_config,
                    node,
                    config,
                )

            self.manifest._parsing_info.static_analysis_parsed_path_count += 1
        # if the static parser didn't succeed, fall back to jinja
        else:
            # jinja rendering
            super().render_update(node, config)
            fire_event(StaticParserFallbackJinjaRendering(path=node.path))

            # if sampling, add the correct messages for tracking
            if exp_sample and isinstance(experimental_sample, str):
                if experimental_sample == "cannot_parse":
                    result += ["01_experimental_parser_cannot_parse"]
                elif experimental_sample == "has_banned_macro":
                    result += ["08_has_banned_macro"]
            elif stable_sample and isinstance(statically_parsed, str):
                if statically_parsed == "cannot_parse":
                    result += ["81_stable_parser_cannot_parse"]
                elif statically_parsed == "has_banned_macro":
                    result += ["88_has_banned_macro"]

        # only send the tracking event if there is at least one result code
        if result:
            # fire a tracking event. this fires one event for every sample
            # so that we have data on a per file basis. Not only can we expect
            # no false positives or misses, we can expect the number model
            # files parseable by the experimental parser to match our internal
            # testing.
            if tracking.active_user is not None:  # None in some tests
                tracking.track_experimental_parser_sample({
                    "project_id": self.root_project.hashed_name(),
                    "file_id": utils.get_hash(node),
                    "status": result
                })

    def run_static_parser(
        self, node: ParsedModelNode
    ) -> Optional[Union[str, Dict[str, List[Any]]]]:
        # if any banned macros have been overridden by the user, we cannot use the static parser.
        if self._has_banned_macro(node):
            # this log line is used for integration testing. If you change
            # the code at the beginning of the line change the tests in
            # test/integration/072_experimental_parser_tests/test_all_experimental_parser.py
            fire_event(StaticParsingMacroOverrideDetected(path=node.path))
            return "has_banned_macro"

        # run the stable static parser and return the results
        try:
            statically_parsed = py_extract_from_source(
                node.raw_sql
            )
            fire_event(StaticParserSuccess(path=node.path))
            return _shift_sources(statically_parsed)
        # if we want information on what features are barring the static
        # parser from reading model files, this is where we would add that
        # since that information is stored in the `ExtractionError`.
        except ExtractionError:
            fire_event(StaticParserFailure(path=node.path))
            return "cannot_parse"

    def run_experimental_parser(
        self, node: ParsedModelNode
    ) -> Optional[Union[str, Dict[str, List[Any]]]]:
        # if any banned macros have been overridden by the user, we cannot use the static parser.
        if self._has_banned_macro(node):
            # this log line is used for integration testing. If you change
            # the code at the beginning of the line change the tests in
            # test/integration/072_experimental_parser_tests/test_all_experimental_parser.py
            fire_event(StaticParsingMacroOverrideDetected(path=node.path))
            return "has_banned_macro"

        # run the experimental parser and return the results
        try:
            # for now, this line calls the stable static parser since there are no
            # experimental features. Change `py_extract_from_source` to the new
            # experimental call when we add additional features.
            experimentally_parsed = py_extract_from_source(
                node.raw_sql
            )
            fire_event(ExperimentalParserSuccess(path=node.path))
            return _shift_sources(experimentally_parsed)
        # if we want information on what features are barring the experimental
        # parser from reading model files, this is where we would add that
        # since that information is stored in the `ExtractionError`.
        except ExtractionError:
            fire_event(ExperimentalParserFailure(path=node.path))
            return "cannot_parse"

    # checks for banned macros
    def _has_banned_macro(
        self, node: ParsedModelNode
    ) -> bool:
        # first check if there is a banned macro defined in scope for this model file
        root_project_name = self.root_project.project_name
        project_name = node.package_name
        banned_macros = ['ref', 'source', 'config']

        all_banned_macro_keys: Iterator[str] = chain.from_iterable(
            map(
                lambda name: [
                    f"macro.{project_name}.{name}",
                    f"macro.{root_project_name}.{name}"
                ],
                banned_macros
            )
        )

        return reduce(
            lambda z, key: z or (key in self.manifest.macros),
            all_banned_macro_keys,
            False
        )

    # this method updates the model node rendered and unrendered config as well
    # as the node object. Used to populate these values when circumventing jinja
    # rendering like the static parser.
    def populate(
        self,
        node: ParsedModelNode,
        config: ContextConfig,
        statically_parsed: Dict[str, Any]
    ):
        # manually fit configs in
        config._config_call_dict = _get_config_call_dict(statically_parsed)

        # if there are hooks present this, it WILL render jinja. Will need to change
        # when the experimental parser supports hooks
        self.update_parsed_node_config(node, config)

        # update the unrendered config with values from the file.
        # values from yaml files are in there already
        node.unrendered_config.update(dict(statically_parsed['configs']))

        # set refs and sources on the node object
        node.refs += statically_parsed['refs']
        node.sources += statically_parsed['sources']

        # configs don't need to be merged into the node because they
        # are read from config._config_call_dict

    # the manifest is often huge so this method avoids deepcopying it
    def partial_deepcopy(self):
        return ModelParser(
            deepcopy(self.project),
            self.manifest,
            deepcopy(self.root_project)
        )


# pure function. safe to use elsewhere, but unlikely to be useful outside this file.
def _get_config_call_dict(
    static_parser_result: Dict[str, Any]
) -> Dict[str, Any]:
    config_call_dict: Dict[str, Any] = {}

    for c in static_parser_result['configs']:
        ContextConfig._add_config_call(config_call_dict, {c[0]: c[1]})

    return config_call_dict


# TODO if we format sources in the extractor to match this type, we won't need this function.
def _shift_sources(
    static_parser_result: Dict[str, List[Any]]
) -> Dict[str, List[Any]]:
    shifted_result = deepcopy(static_parser_result)
    source_calls = []

    for s in static_parser_result['sources']:
        source_calls.append([s[0], s[1]])
    shifted_result['sources'] = source_calls

    return shifted_result


# returns a list of string codes to be sent as a tracking event
def _get_exp_sample_result(
    sample_node: ParsedModelNode,
    sample_config: ContextConfig,
    node: ParsedModelNode,
    config: ContextConfig
) -> List[str]:
    result: List[Tuple[int, str]] = _get_sample_result(sample_node, sample_config, node, config)

    def process(codemsg):
        code, msg = codemsg
        return f"0{code}_experimental_{msg}"

    return list(map(process, result))


# returns a list of string codes to be sent as a tracking event
def _get_stable_sample_result(
    sample_node: ParsedModelNode,
    sample_config: ContextConfig,
    node: ParsedModelNode,
    config: ContextConfig
) -> List[str]:
    result: List[Tuple[int, str]] = _get_sample_result(sample_node, sample_config, node, config)

    def process(codemsg):
        code, msg = codemsg
        return f"8{code}_stable_{msg}"

    return list(map(process, result))


# returns a list of string codes that need a single digit prefix to be prepended
# before being sent as a tracking event
def _get_sample_result(
    sample_node: ParsedModelNode,
    sample_config: ContextConfig,
    node: ParsedModelNode,
    config: ContextConfig
) -> List[Tuple[int, str]]:
    result: List[Tuple[int, str]] = []
    # look for false positive configs
    for k in sample_config._config_call_dict.keys():
        if k not in config._config_call_dict.keys():
            result += [(2, "false_positive_config_value")]
            break

    # look for missed configs
    for k in config._config_call_dict.keys():
        if k not in sample_config._config_call_dict.keys():
            result += [(3, "missed_config_value")]
            break

    # look for false positive sources
    for s in sample_node.sources:
        if s not in node.sources:
            result += [(4, "false_positive_source_value")]
            break

    # look for missed sources
    for s in node.sources:
        if s not in sample_node.sources:
            result += [(5, "missed_source_value")]
            break

    # look for false positive refs
    for r in sample_node.refs:
        if r not in node.refs:
            result += [(6, "false_positive_ref_value")]
            break

    # look for missed refs
    for r in node.refs:
        if r not in sample_node.refs:
            result += [(7, "missed_ref_value")]
            break

    # if there are no errors, return a success value
    if not result:
        result = [(0, "exact_match")]

    return result



import re
import os.path

from dbt.clients.system import run_cmd, rmdir
from dbt.events.functions import fire_event
from dbt.events.types import (
    GitSparseCheckoutSubdirectory, GitProgressCheckoutRevision,
    GitProgressUpdatingExistingDependency, GitProgressPullingNewDependency,
    GitNothingToDo, GitProgressUpdatedCheckoutRange, GitProgressCheckedOutAt
)
from dbt.exceptions import (
    CommandResultError, RuntimeException, bad_package_spec, raise_git_cloning_error,
    raise_git_cloning_problem
)
from packaging import version


def _is_commit(revision: str) -> bool:
    # match SHA-1 git commit
    return bool(re.match(r"\b[0-9a-f]{40}\b", revision))


def _raise_git_cloning_error(repo, revision, error):
    stderr = error.stderr.decode('utf-8').strip()
    if 'usage: git' in stderr:
        stderr = stderr.split('\nusage: git')[0]
    if re.match("fatal: destination path '(.+)' already exists", stderr):
        raise_git_cloning_error(error)

    bad_package_spec(repo, revision, stderr)


def clone(repo, cwd, dirname=None, remove_git_dir=False, revision=None, subdirectory=None):
    has_revision = revision is not None
    is_commit = _is_commit(revision or "")

    clone_cmd = ['git', 'clone', '--depth', '1']
    if subdirectory:
        fire_event(GitSparseCheckoutSubdirectory(subdir=subdirectory))
        out, _ = run_cmd(cwd, ['git', '--version'], env={'LC_ALL': 'C'})
        git_version = version.parse(re.search(r"\d+\.\d+\.\d+", out.decode("utf-8")).group(0))
        if not git_version >= version.parse("2.25.0"):
            # 2.25.0 introduces --sparse
            raise RuntimeError(
                "Please update your git version to pull a dbt package "
                "from a subdirectory: your version is {}, >= 2.25.0 needed".format(git_version)
            )
        clone_cmd.extend(['--filter=blob:none', '--sparse'])

    if has_revision and not is_commit:
        clone_cmd.extend(['--branch', revision])

    clone_cmd.append(repo)

    if dirname is not None:
        clone_cmd.append(dirname)
    try:
        result = run_cmd(cwd, clone_cmd, env={'LC_ALL': 'C'})
    except CommandResultError as exc:
        _raise_git_cloning_error(repo, revision, exc)

    if subdirectory:
        cwd_subdir = os.path.join(cwd, dirname or '')
        clone_cmd_subdir = ['git', 'sparse-checkout', 'set', subdirectory]
        try:
            run_cmd(cwd_subdir, clone_cmd_subdir)
        except CommandResultError as exc:
            _raise_git_cloning_error(repo, revision, exc)

    if remove_git_dir:
        rmdir(os.path.join(dirname, '.git'))

    return result


def list_tags(cwd):
    out, err = run_cmd(cwd, ['git', 'tag', '--list'], env={'LC_ALL': 'C'})
    tags = out.decode('utf-8').strip().split("\n")
    return tags


def _checkout(cwd, repo, revision):
    fire_event(GitProgressCheckoutRevision(revision=revision))

    fetch_cmd = ["git", "fetch", "origin", "--depth", "1"]

    if _is_commit(revision):
        run_cmd(cwd, fetch_cmd + [revision])
    else:
        run_cmd(cwd, ['git', 'remote', 'set-branches', 'origin', revision])
        run_cmd(cwd, fetch_cmd + ["--tags", revision])

    if _is_commit(revision):
        spec = revision
    # Prefer tags to branches if one exists
    elif revision in list_tags(cwd):
        spec = 'tags/{}'.format(revision)
    else:
        spec = 'origin/{}'.format(revision)

    out, err = run_cmd(cwd, ['git', 'reset', '--hard', spec],
                       env={'LC_ALL': 'C'})
    return out, err


def checkout(cwd, repo, revision=None):
    if revision is None:
        revision = 'HEAD'
    try:
        return _checkout(cwd, repo, revision)
    except CommandResultError as exc:
        stderr = exc.stderr.decode('utf-8').strip()
    bad_package_spec(repo, revision, stderr)


def get_current_sha(cwd):
    out, err = run_cmd(cwd, ['git', 'rev-parse', 'HEAD'], env={'LC_ALL': 'C'})

    return out.decode('utf-8')


def remove_remote(cwd):
    return run_cmd(cwd, ['git', 'remote', 'rm', 'origin'], env={'LC_ALL': 'C'})


def clone_and_checkout(repo, cwd, dirname=None, remove_git_dir=False,
                       revision=None, subdirectory=None):
    exists = None
    try:
        _, err = clone(
            repo,
            cwd,
            dirname=dirname,
            remove_git_dir=remove_git_dir,
            subdirectory=subdirectory,
        )
    except CommandResultError as exc:
        err = exc.stderr.decode('utf-8')
        exists = re.match("fatal: destination path '(.+)' already exists", err)
        if not exists:
            raise_git_cloning_problem(repo)

    directory = None
    start_sha = None
    if exists:
        directory = exists.group(1)
        fire_event(GitProgressUpdatingExistingDependency(dir=directory))
    else:
        matches = re.match("Cloning into '(.+)'", err.decode('utf-8'))
        if matches is None:
            raise RuntimeException(
                f'Error cloning {repo} - never saw "Cloning into ..." from git'
            )
        directory = matches.group(1)
        fire_event(GitProgressPullingNewDependency(dir=directory))
    full_path = os.path.join(cwd, directory)
    start_sha = get_current_sha(full_path)
    checkout(full_path, repo, revision)
    end_sha = get_current_sha(full_path)
    if exists:
        if start_sha == end_sha:
            fire_event(GitNothingToDo(sha=start_sha[:7]))
        else:
            fire_event(GitProgressUpdatedCheckoutRange(
                start_sha=start_sha[:7], end_sha=end_sha[:7]
            ))
    else:
        fire_event(GitProgressCheckedOutAt(end_sha=end_sha[:7]))
    return os.path.join(directory, subdirectory or '')



#!/usr/bin/env python
import os
import sys

if sys.version_info < (3, 7):
    print('Error: dbt does not support this version of Python.')
    print('Please upgrade to Python 3.7 or higher.')
    sys.exit(1)


from setuptools import setup
try:
    from setuptools import find_namespace_packages
except ImportError:
    # the user has a downlevel version of setuptools.
    print('Error: dbt requires setuptools v40.1.0 or higher.')
    print('Please upgrade setuptools with "pip install --upgrade setuptools" '
          'and try again')
    sys.exit(1)


this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md')) as f:
    long_description = f.read()


package_name = "dbt-core"
package_version = "1.0.4"
description = """With dbt, data analysts and engineers can build analytics \
the way engineers build applications."""


setup(
    name=package_name,
    version=package_version,
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="dbt Labs",
    author_email="info@dbtlabs.com",
    url="https://github.com/dbt-labs/dbt-core",
    packages=find_namespace_packages(include=['dbt', 'dbt.*']),
    include_package_data = True,
    test_suite='test',
    entry_points={
        'console_scripts': [
            'dbt = dbt.main:main',
        ],
    },
    scripts=[
        'scripts/dbt',
    ],
    install_requires=[
        'Jinja2==2.11.3',
        'MarkupSafe==2.0.1',
        'agate>=1.6,<1.6.4',
        'click>=8,<9',
        'colorama>=0.3.9,<0.4.5',
        'hologram==0.0.14',
        'isodate>=0.6,<0.7',
        'logbook>=1.5,<1.6',
        'mashumaro==2.9',
        'minimal-snowplow-tracker==0.0.2',
        'networkx>=2.3,<3',
        'packaging>=20.9,<22.0',
        'sqlparse>=0.2.3,<0.5',
        'dbt-extractor~=0.4.0',
        'typing-extensions>=3.7.4,<3.11',
        'werkzeug>=1,<3',
        # the following are all to match snowflake-connector-python
        'requests<3.0.0',
        'idna>=2.5,<4',
        'cffi>=1.9,<2.0.0',
    ],
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'License :: OSI Approved :: Apache Software License',

        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    python_requires=">=3.7",
)



from typing import Dict, Any, Tuple, Optional, Union, Callable

from dbt.clients.jinja import get_rendered, catch_jinja
from dbt.context.target import TargetContext
from dbt.context.secret import SecretContext
from dbt.context.base import BaseContext
from dbt.contracts.connection import HasCredentials
from dbt.exceptions import (
    DbtProjectError, CompilationException, RecursionException
)
from dbt.utils import deep_map_render


Keypath = Tuple[Union[str, int], ...]


class BaseRenderer:
    def __init__(self, context: Dict[str, Any]) -> None:
        self.context = context

    @property
    def name(self):
        return 'Rendering'

    def should_render_keypath(self, keypath: Keypath) -> bool:
        return True

    def render_entry(self, value: Any, keypath: Keypath) -> Any:
        if not self.should_render_keypath(keypath):
            return value

        return self.render_value(value, keypath)

    def render_value(
        self, value: Any, keypath: Optional[Keypath] = None
    ) -> Any:
        # keypath is ignored.
        # if it wasn't read as a string, ignore it
        if not isinstance(value, str):
            return value
        try:
            with catch_jinja():
                return get_rendered(value, self.context, native=True)
        except CompilationException as exc:
            msg = f'Could not render {value}: {exc.msg}'
            raise CompilationException(msg) from exc

    def render_data(
        self, data: Dict[str, Any]
    ) -> Dict[str, Any]:
        try:
            return deep_map_render(self.render_entry, data)
        except RecursionException:
            raise DbtProjectError(
                f'Cycle detected: {self.name} input has a reference to itself',
                project=data
            )


def _list_if_none(value):
    if value is None:
        value = []
    return value


def _dict_if_none(value):
    if value is None:
        value = {}
    return value


def _list_if_none_or_string(value):
    value = _list_if_none(value)
    if isinstance(value, str):
        return [value]
    return value


class ProjectPostprocessor(Dict[Keypath, Callable[[Any], Any]]):
    def __init__(self):
        super().__init__()

        self[('on-run-start',)] = _list_if_none_or_string
        self[('on-run-end',)] = _list_if_none_or_string

        for k in ('models', 'seeds', 'snapshots'):
            self[(k,)] = _dict_if_none
            self[(k, 'vars')] = _dict_if_none
            self[(k, 'pre-hook')] = _list_if_none_or_string
            self[(k, 'post-hook')] = _list_if_none_or_string
        self[('seeds', 'column_types')] = _dict_if_none

    def postprocess(self, value: Any, key: Keypath) -> Any:
        if key in self:
            handler = self[key]
            return handler(value)

        return value


class DbtProjectYamlRenderer(BaseRenderer):
    _KEYPATH_HANDLERS = ProjectPostprocessor()

    def __init__(
        self, profile: Optional[HasCredentials] = None,
        cli_vars: Optional[Dict[str, Any]] = None
    ) -> None:
        # Generate contexts here because we want to save the context
        # object in order to retrieve the env_vars. This is almost always
        # a TargetContext, but in the debug task we want a project
        # even when we don't have a profile.
        if cli_vars is None:
            cli_vars = {}
        if profile:
            self.ctx_obj = TargetContext(profile, cli_vars)
        else:
            self.ctx_obj = BaseContext(cli_vars)  # type:ignore
        context = self.ctx_obj.to_dict()
        super().__init__(context)

    @property
    def name(self):
        'Project config'

    def get_package_renderer(self) -> BaseRenderer:
        return PackageRenderer(self.context)

    def get_selector_renderer(self) -> BaseRenderer:
        return SelectorRenderer(self.context)

    def render_project(
        self,
        project: Dict[str, Any],
        project_root: str,
    ) -> Dict[str, Any]:
        """Render the project and insert the project root after rendering."""
        rendered_project = self.render_data(project)
        rendered_project['project-root'] = project_root
        return rendered_project

    def render_packages(self, packages: Dict[str, Any]):
        """Render the given packages dict"""
        package_renderer = self.get_package_renderer()
        return package_renderer.render_data(packages)

    def render_selectors(self, selectors: Dict[str, Any]):
        selector_renderer = self.get_selector_renderer()
        return selector_renderer.render_data(selectors)

    def render_entry(self, value: Any, keypath: Keypath) -> Any:
        result = super().render_entry(value, keypath)
        return self._KEYPATH_HANDLERS.postprocess(result, keypath)

    def should_render_keypath(self, keypath: Keypath) -> bool:
        if not keypath:
            return True

        first = keypath[0]
        # run hooks are not rendered
        if first in {'on-run-start', 'on-run-end', 'query-comment'}:
            return False

        # don't render vars blocks until runtime
        if first == 'vars':
            return False

        if first in {'seeds', 'models', 'snapshots', 'tests'}:
            keypath_parts = {
                (k.lstrip('+ ') if isinstance(k, str) else k)
                for k in keypath
            }
            # model-level hooks
            if 'pre-hook' in keypath_parts or 'post-hook' in keypath_parts:
                return False

        return True


class SelectorRenderer(BaseRenderer):
    @property
    def name(self):
        return 'Selector config'


class SecretRenderer(BaseRenderer):
    def __init__(
        self, cli_vars: Optional[Dict[str, Any]] = None
    ) -> None:
        # Generate contexts here because we want to save the context
        # object in order to retrieve the env_vars.
        if cli_vars is None:
            cli_vars = {}
        self.ctx_obj = SecretContext(cli_vars)
        context = self.ctx_obj.to_dict()
        super().__init__(context)

    @property
    def name(self):
        return 'Secret'


class ProfileRenderer(SecretRenderer):
    @property
    def name(self):
        return 'Profile'


class PackageRenderer(SecretRenderer):
    @property
    def name(self):
        return 'Packages config'



from pathlib import Path
from .graph.manifest import WritableManifest
from .results import RunResultsArtifact
from typing import Optional
from dbt.exceptions import IncompatibleSchemaException


class PreviousState:
    def __init__(self, path: Path):
        self.path: Path = path
        self.manifest: Optional[WritableManifest] = None
        self.results: Optional[RunResultsArtifact] = None

        manifest_path = self.path / 'manifest.json'
        if manifest_path.exists() and manifest_path.is_file():
            try:
                # we want to bail with an error if schema versions don't match
                self.manifest = WritableManifest.read_and_check_versions(str(manifest_path))
            except IncompatibleSchemaException as exc:
                exc.add_filename(str(manifest_path))
                raise

        results_path = self.path / 'run_results.json'
        if results_path.exists() and results_path.is_file():
            try:
                # we want to bail with an error if schema versions don't match
                self.results = RunResultsArtifact.read_and_check_versions(str(results_path))
            except IncompatibleSchemaException as exc:
                exc.add_filename(str(results_path))
                raise
